#! /usr/bin/env python

import itertools

from . import common
from . import core
from .. import html401 as html
from .. import imsmdv1p2p1 as imsmd
from ..py2 import is_text
from ..qtiv2 import xml as qtiv2
from ..xml import structures as xml
from ..xml import xsdatatypes as xsi


class Item(
        common.QTICommentContainer,
        core.SectionItemMixin,
        core.ObjectMixin):

    """The Item is the smallest unit data structure that can be
    exchanged using the QTI specification. Each Item consists of five
    distinct parts, namely: objectives - the materials used to describe
    the objectives with respect to each view; rubric - the materials
    used to define the context of the Item and available for each view;
    presentation - the instructions describing the nature of the
    question to be asked; resprocessing - the instructions to be
    followed when analyzing the responses to create a corresponding
    score and feedback; itemfeedback - the materials to be presented as
    feedback to the entered response::

    <!ELEMENT item (qticomment?
            duration?
            itemmetadata?
            objectives*
            itemcontrol*
            itemprecondition*
            itempostcondition*
            (itemrubric | rubric)*
            presentation?
            resprocessing*
            itemproc_extension?
            itemfeedback*
            reference?)>
    <!ATTLIST item  maxattempts CDATA  #IMPLIED
            label CDATA  #IMPLIED
            ident CDATA  #REQUIRED
            title CDATA  #IMPLIED
            xml:lang    CDATA  #IMPLIED >"""
    XMLNAME = 'item'
    XMLATTR_maxattempts = 'maxattempts'
    XMLATTR_label = 'label'
    XMLATTR_ident = 'ident'
    XMLATTR_title = 'title'
    XMLCONTENT = xml.ElementContent

    def __init__(self, parent):
        common.QTICommentContainer.__init__(self, parent)
        self.maxattempts = None
        self.label = None
        self.ident = None
        self.title = None
        self.Duration = None
        self.ItemMetadata = None
        self.Objectives = []
        self.ItemControl = []
        self.ItemPreCondition = []
        self.ItemPostCondition = []
        self.Rubric = []				#: includes ItemRubric
        self.Presentation = None
        self.ResProcessing = []
        self.ItemProcExtension = None
        self.ItemFeedback = []
        self.QTIReference = None

    def get_children(self):
        for child in common.QTICommentContainer.get_children(self):
            yield child
        if self.Duration:
            yield self.Duration
        if self.ItemMetadata:
            yield self.ItemMetadata
        for child in itertools.chain(
                self.Objectives,
                self.ItemControl,
                self.ItemPreCondition,
                self.ItemPostCondition,
                self.Rubric):
            yield child
        if self.Presentation:
            yield self.Presentation
        for child in self.ResProcessing:
            yield child
        if self.ItemProcExtension:
            yield self.ItemProcExtension
        for child in self.ItemFeedback:
            yield child
        if self.QTIReference:
            yield self.QTIReference

    def migrate_to_v2(self, output):
        """Converts this item to QTI v2

        For details, see
        :py:meth:`pyslet.qtiv1.QuesTestInterop.migrate_to_v2`."""
        # First thing we do is initialize any fixups
        for rp in self.ResProcessing:
            rp._interactionFixup = {}
        doc = qtiv2.QTIDocument(root=qtiv2.items.AssessmentItem)
        item = doc.root
        lom = imsmd.LOM(None)
        log = []
        ident = qtiv2.core.ValidateIdentifier(self.ident)
        if self.ident != ident:
            log.append(
                "Warning: illegal NCName for ident: %s, replaced with: %s" %
                (self.ident, ident))
        item.identifier = ident
        title = self.title
        # may be specified in the metadata
        if self.ItemMetadata:
            md_titles = self.ItemMetadata.metadata.get('title', ())
        else:
            md_titles = ()
        if title:
            item.title = title
        elif md_titles:
            item.title = md_titles[0][0]
        else:
            item.title = ident
        if self.maxattempts is not None:
            log.append(
                "Warning: maxattempts can not be controlled at item level,"
                " ignored: maxattempts='" +
                self.maxattempts +
                "'")
        if self.label:
            item.label = self.label
        lang = self.resolve_lang()
        item.set_lang(lang)
        general = lom.add_child(imsmd.LOMGeneral)
        id = general.add_child(imsmd.LOMIdentifier)
        id.set_value(self.ident)
        if title:
            lom_title = general.add_child(imsmd.LOMTitle)
            lom_title = lom_title.add_child(lom_title.LangStringClass)
            lom_title.set_value(title)
            if lang:
                lom_title.set_lang(lang)
        if md_titles:
            if title:
                # If we already have a title, then we have to add
                # qmd_title as description metadata you may think
                # qmd_title is a better choice than the title attribute
                # but qmd_title is an extension so the title attribute
                # takes precedence
                i = 0
            else:
                lom_title = general.add_child(imsmd.LOMTitle)
                lom_title = lom_title.add_child(lom_title.LangStringClass)
                lom_title.set_value(md_titles[0][0])
                lang = md_titles[0][1].resolve_lang()
                if lang:
                    lom_title.set_lang(lang)
                i = 1
            for mdTitle in md_titles[i:]:
                description = general.add_child(general.DescriptionClass)
                lom_title = description.add_child(
                    description.LangStringClass)
                lom_title.set_value(mdTitle[0])
                md_lang = mdTitle[1].resolve_lang()
                if md_lang:
                    lom_title.set_lang(md_lang)
        if self.QTIComment:
            # A comment on an item is added as a description to the metadata
            description = general.add_child(general.DescriptionClass)
            description.add_child(description.LangStringClass).set_value(
                self.QTIComment.get_value())
        if self.Duration:
            log.append(
                "Warning: duration is currently outside the scope of"
                " version 2: ignored " +
                self.Duration.get_value())
        if self.ItemMetadata:
            self.ItemMetadata.migrate_to_v2(doc, lom, log)
        for objective in self.Objectives:
            if objective.view != core.View.All:
                objective.migrate_to_v2(item, log)
            else:
                objective.migrate_objectives_to_lrm(lom, log)
        if self.ItemControl:
            log.append(
                "Warning: itemcontrol is currently outside the scope of"
                " version 2")
        for rubric in self.Rubric:
            rubric.migrate_to_v2(item, log)
        if self.Presentation:
            self.Presentation.migrate_to_v2(item, log)
        if self.ResProcessing:
            if len(self.ResProcessing) > 1:
                log.append(
                    "Warning: multiople <resprocessing> not supported,"
                    " ignoring all but the last")
            self.ResProcessing[-1].migrate_to_v2(item, log)
        for feedback in self.ItemFeedback:
            feedback.migrate_to_v2(item, log)
        item.sort_declarations()
        output.append((doc, lom, log))
        # print doc.root


class ItemMetadata(common.MetadataContainerMixin, core.QTIElement):

    """The itemmetadata element contains all of the QTI-specific
    meta-data to be applied to the Item. This meta-data can consist of
    either entries defined using an external vocabulary or the
    individually named entries::

    <!ELEMENT itemmetadata (
            qtimetadata*
            qmd_computerscored?
            qmd_feedbackpermitted?
            qmd_hintspermitted?
            qmd_itemtype?
            qmd_levelofdifficulty?
            qmd_maximumscore?
            qmd_renderingtype*
            qmd_responsetype*
            qmd_scoringpermitted?
            qmd_solutionspermitted?
            qmd_status?
            qmd_timedependence?
            qmd_timelimit?
            qmd_toolvendor?
            qmd_topic?
            qmd_weighting?
            qmd_material*
            qmd_typeofsolution?
            )>

    This element contains more structure than is in common use, at the
    moment we represent this structure directly and automaticaly conform
    output to it, adding extension elements at the end.  In the future
    we might be more generous and allow input *and* output of elements
    in any sequence and provide separate methods for conforming these
    elements."""
    XMLNAME = 'itemmetadata'
    XMLCONTENT = xml.ElementContent

    def __init__(self, parent):
        core.QTIElement.__init__(self, parent)
        common.MetadataContainerMixin.__init__(self)
        self.QTIMetadata = []
        self.QMDComputerScored = None
        self.QMDFeedbackPermitted = None
        self.QMDHintsPermitted = None
        self.QMDItemType = None
        self.QMDLevelOfDifficulty = None
        self.QMDMaximumScore = None
        self.QMDRenderingType = []
        self.QMDResponseType = []
        self.QMDScoringPermitted = None
        self.QMDSolutionsPermitted = None
        self.QMDStatus = None
        self.QMDTimeDependence = None
        self.QMDTimeLimit = None
        self.QMDToolVendor = None
        self.QMDTopic = None
        self.QMDWeighting = None
        self.QMDMaterial = []
        self.QMDTypeOfSolution = None
        # Extensions in common use....
        self.QMDAuthor = []
        self.QMDDescription = []
        self.QMDDomain = []
        self.QMDKeywords = []
        self.QMDOrganization = []
        self.QMDTitle = None

    def get_children(self):
        for child in self.QTIMetadata:
            yield child
        if self.QMDComputerScored:
            yield self.QMDComputerScored
        if self.QMDFeedbackPermitted:
            yield self.QMDFeedbackPermitted
        if self.QMDHintsPermitted:
            yield self.QMDHintsPermitted
        if self.QMDItemType:
            yield self.QMDItemType
        if self.QMDLevelOfDifficulty:
            yield self.QMDLevelOfDifficulty
        if self.QMDMaximumScore:
            yield self.QMDMaximumScore
        for child in itertools.chain(
                self.QMDRenderingType,
                self.QMDResponseType):
            yield child
        if self.QMDScoringPermitted:
            yield self.QMDScoringPermitted
        if self.QMDSolutionsPermitted:
            yield self.QMDSolutionsPermitted
        if self.QMDStatus:
            yield self.QMDStatus
        if self.QMDTimeDependence:
            yield self.QMDTimeDependence
        if self.QMDTimeLimit:
            yield self.QMDTimeLimit
        if self.QMDToolVendor:
            yield self.QMDToolVendor
        if self.QMDTopic:
            yield self.QMDTopic
        if self.QMDWeighting:
            yield self.QMDWeighting
        for child in self.QMDMaterial:
            yield child
        if self.QMDTypeOfSolution:
            yield self.QMDTypeOfSolution
        for child in itertools.chain(
                self.QMDAuthor,
                self.QMDDescription,
                self.QMDDomain,
                self.QMDKeywords,
                self.QMDOrganization):
            yield child
        if self.QMDTitle:
            yield self.QMDTitle
        for child in core.QTIElement.get_children(self):
            yield child

    def migrate_level_of_difficulty_to_lrm(self, lom, log):
        difficulty = self.metadata.get('levelofdifficulty', ())
        for value, definition in difficulty:
            # IMS Definition says: The options are: "Pre-school",
            # "School" or "HE/FE", "Vocational" and "Professional
            # Development" so we bind this value to the "Context" in LOM
            # if one of the QTI or LOM defined terms have been used,
            # otherwise, we bind to Difficulty, as this seems to be more
            # common usage.
            context, lom_flag = QMDLevelOfDifficulty.LOMContextMap.get(
                value.lower(), (None, False))
            educational = lom.add_child(imsmd.LOMEducational)
            if context is None:
                # add value as difficulty
                lom_flag = value.lower(
                ) in QMDLevelOfDifficulty.LOMDifficultyMap
                d = educational.add_child(imsmd.LOMDifficulty)
                if lom_flag:
                    d.LRMSource.LangString.set_value(imsmd.LOM_SOURCE)
                else:
                    d.LRMSource.LangString.set_value(imsmd.LOM_UNKNOWNSOURCE)
                d.LRMSource.LangString.set_lang("x-none")
                d.LRMValue.LangString.set_value(value)
                d.LRMValue.LangString.set_lang("x-none")
            else:
                # add value as educational context
                c = educational.add_child(imsmd.LOMContext)
                if lom_flag:
                    c.LRMSource.LangString.set_value(imsmd.LOM_SOURCE)
                else:
                    c.LRMSource.LangString.set_value(imsmd.LOM_UNKNOWNSOURCE)
                c.LRMSource.LangString.set_lang("x-none")
                c.LRMValue.LangString.set_value(context)
                c.LRMValue.LangString.set_lang("x-none")

    def migrate_status_to_lrm(self, lom, log):
        status = self.metadata.get('status', ())
        for value, definition in status:
            s = lom.add_child(
                imsmd.LOMLifecycle).add_child(imsmd.LOMStatus)
            value = value.lower()
            source = QMDStatus.SourceMap.get(value, imsmd.LOM_UNKNOWNSOURCE)
            s.LRMSource.LangString.set_value(source)
            s.LRMSource.LangString.set_lang("x-none")
            s.LRMValue.LangString.set_value(value)
            s.LRMValue.LangString.set_lang("x-none")

    def migrate_topic_to_lrm(self, lom, log):
        topics = self.metadata.get('topic', ())
        for value, definition in topics:
            lang = definition.resolve_lang()
            value = value.strip()
            description = lom.add_child(
                imsmd.LOMEducational).add_child(imsmd.Description)
            description.AddString(lang, value)

    def migrate_contributor_to_lrm(self, field_name, lom_role, lom, log):
        contributors = self.metadata.get(field_name, ())
        if contributors:
            if imsmd.vobject is None:
                log.append(
                    'Warning: qmd_%s support disabled'
                    ' (vobject not installed)' % field_name)
            else:
                for value, definition in contributors:
                    lifecycle = lom.add_child(imsmd.LOMLifecycle)
                    contributor = lifecycle.add_child(imsmd.LOMContribute)
                    role = contributor.LOMRole
                    role.LRMSource.LangString.set_value(imsmd.LOM_SOURCE)
                    role.LRMSource.LangString.set_lang("x-none")
                    role.LRMValue.LangString.set_value(lom_role)
                    role.LRMValue.LangString.set_lang("x-none")
                    names = value.strip().split(',')
                    for name in names:
                        if not name.strip():
                            continue
                        vcard = imsmd.vobject.vCard()
                        vcard.add('n')
                        vcard.n.value = imsmd.vobject.vcard.Name(
                            family=name, given='')
                        vcard.add('fn')
                        vcard.fn.value = name.strip()
                        contributor.add_child(
                            imsmd.LOMCEntity).LOMVCard.set_value(vcard)

    def migrate_description_to_lrm(self, lom, log):
        descriptions = self.metadata.get('description', ())
        for value, definition in descriptions:
            lang = definition.resolve_lang()
            gen_description = lom.add_child(
                imsmd.LOMGeneral).add_child(imsmd.Description)
            gen_description = gen_description.add_child(
                gen_description.LangStringClass)
            gen_description.set_value(value)
            if lang:
                gen_description.set_lang(lang)

    def migrate_domain_to_lrm(self, lom, log):
        domains = self.metadata.get('domain', ())
        warn = False
        for value, definition in domains:
            lang = definition.resolve_lang()
            kw_value = value.strip()
            if kw_value:
                kw_container = lom.add_child(
                    imsmd.LOMGeneral).add_child(imsmd.LOMKeyword)
                kw_container = kw_container.add_child(
                    kw_container.LangStringClass)
                kw_container.set_value(kw_value)
                # set the language of the kw
                if lang:
                    kw_container.set_lang(lang)
                if not warn:
                    log.append(
                        "Warning: qmd_domain extension field will be added"
                        " as LOM keyword")
                    warn = True

    def migrate_keywords_to_lrm(self, lom, log):
        keywords = self.metadata.get('keywords', ())
        for value, definition in keywords:
            lang = definition.resolve_lang()
            values = value.split(',')
            for kw_value in values:
                v = kw_value.strip()
                if v:
                    kw_container = lom.add_child(
                        imsmd.LOMGeneral).add_child(imsmd.LOMKeyword)
                    kw_container = kw_container.add_child(
                        kw_container.LangStringClass)
                    kw_container.set_value(v)
                    # set the language of the kw
                    if lang:
                        kw_container.set_lang(lang)

    def migrate_organization_to_lrm(self, lom, log):
        organizations = self.metadata.get('organization', ())
        if organizations:
            if imsmd.vobject is None:
                log.append(
                    'Warning: qmd_organization support disabled'
                    ' (vobject not installed)')
            else:
                for value, definition in organizations:
                    lifecycle = lom.add_child(imsmd.LOMLifecycle)
                    contributor = lifecycle.add_child(imsmd.LOMContribute)
                    role = contributor.LOMRole
                    role.LRMSource.LangString.set_value(imsmd.LOM_SOURCE)
                    role.LRMSource.LangString.set_lang("x-none")
                    role.LRMValue.LangString.set_value("unknown")
                    role.LRMValue.LangString.set_lang("x-none")
                    name = value.strip()
                    vcard = imsmd.vobject.vCard()
                    vcard.add('n')
                    vcard.n.value = imsmd.vobject.vcard.Name(
                        family=name, given='')
                    vcard.add('fn')
                    vcard.fn.value = name
                    vcard.add('org')
                    vcard.org.value = [name]
                    contributor.add_child(
                        imsmd.LOMCEntity).LOMVCard.set_value(vcard)

    def migrate_to_v2(self, doc, lom, log):
        item = doc.root
        itemtypes = self.metadata.get('itemtype', ())
        for itemtype, itemtypeDef in itemtypes:
            log.append(
                "Warning: qmd_itemtype now replaced by"
                " qtiMetadata.interactionType in manifest, ignoring %s" %
                itemtype)
        self.migrate_level_of_difficulty_to_lrm(lom, log)
        self.migrate_status_to_lrm(lom, log)
        vendors = self.metadata.get('toolvendor', ())
        for value, definition in vendors:
            item.metadata.add_child(qtiv2.metadata.ToolVendor).set_value(value)
        self.migrate_topic_to_lrm(lom, log)
        self.migrate_contributor_to_lrm('author', 'author', lom, log)
        self.migrate_contributor_to_lrm('creator', 'initiator', lom, log)
        self.migrate_contributor_to_lrm('owner', 'publisher', lom, log)
        self.migrate_description_to_lrm(lom, log)
        self.migrate_domain_to_lrm(lom, log)
        self.migrate_keywords_to_lrm(lom, log)
        self.migrate_organization_to_lrm(lom, log)


class QMDMetadataElement(core.QTIElement):

    """Abstract class to represent old-style qmd_ tags"""

    def content_changed(self):
        self.declare_metadata(self.get_xmlname(), self.get_value(), self)


class QMDComputerScored(QMDMetadataElement):

    """Whether or not the Item can be computer scored."""
    XMLNAME = 'qmd_computerscored'


class QMDFeedbackPermitted(QMDMetadataElement):

    """Indicates whether or not feedback is available within the Item."""
    XMLNAME = 'qmd_feedbackpermitted'


class QMDHintsPermitted(QMDMetadataElement):

    """Indicates whether or not hints are available within the Item."""
    XMLNAME = 'qmd_hintspermitted'


class QMDItemType(QMDMetadataElement):

    """The type of Item available."""
    XMLNAME = 'qmd_itemtype'


class QMDLevelOfDifficulty(QMDMetadataElement):

    """The educational level for which the Item is intended::

    <!ELEMENT qmd_levelofdifficulty (#PCDATA)>"""
    XMLNAME = 'qmd_levelofdifficulty'

    LOMDifficultyMap = {
        "very easy": True,
        "easy": True,
        "medium": True,
        "difficult": True,
        "very difficult": True
    }
    """A mapping from difficulty values to the LOM difficulties.  This
    value is supposed to be an educational level, e.g., "pre-school" but
    in practice it is often used for values more akin to the LOM concept
    of difficulty."""

    LOMContextMap = {
        # value is outside LOM defined vocab
        "pre-school": ("pre-school", False),
        "school": ("school", True),
        "he/fe": ("higher education", True),
        # value is outside LOM defined vocab
        "vocational": ("vocational", False),
        "professional development": ("training", True)
    }
    """A mapping from difficulty values to the LOM contexts.  This mapping
    returns a tuple of the context value and a boolean indicating whether or
    not this is one from the LOM standard or not."""


class QMDMaterial(QMDMetadataElement):

    """The type of material content used within the Item."""
    XMLNAME = 'qmd_material'


class QMDMaximumScore(QMDMetadataElement):

    """The maximum score possible from the Item."""
    XMLNAME = 'qmd_maximumscore'


class QMDRenderingType(QMDMetadataElement):

    """The type of rendering used within the Item."""
    XMLNAME = 'qmd_renderingtype'


class QMDResponseType(QMDMetadataElement):

    """The class of response expected for the Item."""
    XMLNAME = 'qmd_responsetype'


class QMDScoringPermitted(QMDMetadataElement):

    """Indicates whether or not scoring is available within the Item."""
    XMLNAME = 'qmd_scoringpermitted'


class QMDSolutionsPermitted(QMDMetadataElement):

    """Indicates whether or not solutions are available within the Item."""
    XMLNAME = 'qmd_solutionspermitted'


class QMDStatus(QMDMetadataElement):

    """The status of the Item."""
    XMLNAME = 'qmd_status'

    SourceMap = {
        'draft': imsmd.LOM_SOURCE,
        'final': imsmd.LOM_SOURCE,
        'revised': imsmd.LOM_SOURCE,
        'unavailable': imsmd.LOM_SOURCE,
        'experimental': core.QTI_SOURCE,
        'normal': core.QTI_SOURCE,
        'retired': core.QTI_SOURCE
    }
    """A mapping from status values to a metadata source, one of the
    following constants: :py:data:`~pyslet.imsmdv1p2p1.LOM_SOURCE`
    :py:data:`~pyslet.qtiv1.core.QTI_SOURCE`"""


class QMDTimeDependence(QMDMetadataElement):

    """Whether or not the responses are time dependent."""
    XMLNAME = 'qmd_timedependence'


class QMDTimeLimit(QMDMetadataElement):

    """The number of minutes to be permitted for the completion of the Item."""
    XMLNAME = 'qmd_timelimit'


class QMDToolVendor(QMDMetadataElement):

    """The name of the vendor of the tool creating the Item."""
    XMLNAME = 'qmd_toolvendor'


class QMDTopic(QMDMetadataElement):

    """The brief description of the topic covered by the Item."""
    XMLNAME = 'qmd_topic'


class QMDTypeOfSolution(QMDMetadataElement):

    """The type of solution available in the Item."""
    XMLNAME = 'qmd_typeofsolution'


class QMDWeighting(QMDMetadataElement):

    """The weighting to be applied to the scores allocated to this Item
    before aggregations with other scores."""
    XMLNAME = 'qmd_weighting'


class QMDAuthor(QMDMetadataElement):

    """Not defined by QTI but seems to be in common use."""
    XMLNAME = 'qmd_author'


class QMDDescription(QMDMetadataElement):

    """Not defined by QTI but seems to be in common use."""
    XMLNAME = 'qmd_description'


class QMDDomain(QMDMetadataElement):

    """Not defined by QTI but seems to be in common use."""
    XMLNAME = 'qmd_domain'


class QMDKeywords(QMDMetadataElement):

    """Not defined by QTI but seems to be in common use."""
    XMLNAME = 'qmd_keywords'


class QMDOrganization(QMDMetadataElement):

    """Not defined by QTI but seems to be in common use."""
    XMLNAME = 'qmd_organization'


class QMDTitle(QMDMetadataElement):

    """Not defined by QTI but seems to be in common use."""
    XMLNAME = 'qmd_title'


class ItemPreCondition(core.QTIElement):

    """The preconditions that control whether or not the Item is activated::

    <!ELEMENT itemprecondition (#PCDATA)>"""
    XMLNAME = 'itemprecondition'
    XMLCONTENT = xml.XMLMixedContent


class ItemPostCondition(core.QTIElement):

    """The postconditions that control whether or not the next Item may be
    activated::

    <!ELEMENT itempostcondition (#PCDATA)>"""
    XMLNAME = 'itempostcondition'
    XMLCONTENT = xml.XMLMixedContent


class ItemControl(common.QTICommentContainer):

    """The control switches that are used to enable or disable the display of
    hints, solutions and feedback within the Item::

    <!ELEMENT itemcontrol (qticomment?)>
    <!ATTLIST itemcontrol  feedbackswitch  (Yes | No )  'Yes'
            hintswitch  (Yes | No )  'Yes'
            solutionswitch  (Yes | No )  'Yes'
            view	(All | Administrator | AdminAuthority | Assessor | Author |
                            Candidate | InvigilatorProctor | Psychometrician |
                            Scorer | Tutor ) 'All' >
    """
    XMLNAME = 'itemcontrol'
    XMLATTR_feedbackswitch = (
        'feedbackSwitch', core.ParseYesNo, core.FormatYesNo)
    XMLATTR_hintswitch = ('hintSwitch', core.ParseYesNo, core.FormatYesNo)
    XMLATTR_solutionswitch = (
        'solutionSwitch', core.ParseYesNo, core.FormatYesNo)
    XMLCONTENT = xml.ElementContent

    def __init__(self, parent):
        common.QTICommentContainer.__init__(self, parent)
        self.feedbackSwitch = True
        self.hintSwitch = True
        self.solutionSwitch = True
        self.view = core.View.DEFAULT


class ItemRubric(common.Rubric):

    """The itemrubric element is used to contain contextual information
    that is important to the element e.g. it could contain standard data
    values that might or might not be useful for answering the question.
    Different sets of rubric can be defined for each of the possible
    views::

    <!ELEMENT itemrubric (material)>
    <!ATTLIST itemrubric
            view	(All | Administrator | AdminAuthority | Assessor |
                     Author | Candidate | InvigilatorProctor |
                     Psychometrician | Scorer | Tutor ) 'All' >

    We are generous with this element, extending the allowable content
    model to make it equivalent to <rubric> which is a superset.
    <itemrubric> was deprecated in favour of <rubric> with QTI v1.2"""
    XMLNAME = 'itemrubric'
    XMLCONTENT = xml.ElementContent


class FlowContainer(common.ContentMixin, common.QTICommentContainer):

    """Abstract class used to represent elements that contain flow and
    related elements::

    <!ELEMENT XXXXXXXXXX (qticomment? , (material | flow | response_*)* )>"""

    def __init__(self, parent):
        common.QTICommentContainer.__init__(self, parent)
        common.ContentMixin.__init__(self)

    def get_children(self):
        return itertools.chain(
            common.QTICommentContainer.get_children(self),
            common.ContentMixin.get_content_children(self))

    def content_child(self, child_class):
        return child_class in (common.Material, Flow) or \
            issubclass(child_class, Response)


class Presentation(FlowContainer, common.PositionMixin):

    """This element contains all of the instructions for the
    presentation of the question during an evaluation. This information
    includes the actual material to be presented. The labels for the
    possible responses are also identified and these are used by the
    response processing element defined elsewhere in the Item::

    <!ELEMENT presentation (qticomment? ,
            (flow |
                    (material |
                    response_lid |
                    response_xy |
                    response_str |
                    response_num |
                    response_grp |
                    response_extension)+
                    )
            )>
    <!ATTLIST presentation  label CDATA  #IMPLIED
            xml:lang CDATA  #IMPLIED
            y0 CDATA  #IMPLIED
            x0 CDATA  #IMPLIED
            width CDATA  #IMPLIED
            height CDATA  #IMPLIED >"""
    XMLNAME = 'presentation'
    XMLATTR_label = 'label'

    XMLCONTENT = xml.ElementContent

    def __init__(self, parent):
        FlowContainer.__init__(self, parent)
        common.PositionMixin.__init__(self)
        self.label = None

    def migrate_to_v2(self, v2_item, log):
        """Presentation maps to the main content in item_body."""
        item_body = v2_item.get_or_add_child(qtiv2.content.ItemBody)
        if self.got_position():
            log.append(
                "Warning: discarding absolute positioning information"
                " on presentation")
        if self.inline_children():
            p = item_body.add_child(
                html.P, (qtiv2.core.IMSQTI_NAMESPACE, 'p'))
            if self.label is not None:
                # p.label=self.label
                p.set_attribute('label', self.label)
            self.migrate_content_to_v2(p, html.InlineMixin, log)
        elif self.label is not None:
            # We must generate a div to hold the label, we can't rely on
            # owning the whole item_body
            div = item_body.add_child(
                html.Div, (qtiv2.core.IMSQTI_NAMESPACE, 'div'))
            div.set_attribute('label', self.label)
            # Although div will take an inline directly we force blocking at
            # the top level
            self.migrate_content_to_v2(div, html.BlockMixin, log)
        else:
            # mixture or block children, force use of blocks
            self.migrate_content_to_v2(item_body, html.BlockMixin, log)
        self.clean_hotspot_images(item_body)

    def clean_hotspot_images(self, item_body):
        """Removes spurious img tags which represent images used in
        hotspotInteractions.

        Unfortunately we have to do this because images needed in hotspot
        interactions are often clumsily placed outside the response/render
        constructs.  Rather than fiddle around at the time we simply migrate
        the lot, duplicating the images in the hotspotInteractions.  When the
        item_body is complete we do a grand tidy up to remove spurious
        images."""
        hotspots = list(
            item_body.find_children_depth_first(
                (qtiv2.interactions.HotspotInteraction,
                 qtiv2.interactions.SelectPointInteraction),
                False))
        images = list(item_body.find_children_depth_first(html.Img, False))
        for hs in hotspots:
            for img in images:
                # migrated images/hotspots will always have absolute URIs
                if img.src and str(img.src) == str(hs.Object.data):
                    parent = img.parent
                    parent.remove_child(img)
                    if (isinstance(parent, html.P) and
                            len(list(parent.get_children())) == 0):
                        # It is always safe to remove a paragraph left
                        # empty by deleting an image. The chances are
                        # the paragraph was created by us to house a
                        # matimage
                        parent.parent.remove_child(parent)

    def is_inline(self):
        return False


class Flow(FlowContainer, common.FlowMixin):

    """This element contains all of the instructions for the
    presentation with flow blocking of the question during a test. This
    information includes the actual material to be presented. The labels
    for the possible responses are also identified and these are used by
    the response processing element defined elsewhere in the Item::

    <!ELEMENT flow (qticomment? ,
            (flow |
            material |
            material_ref |
            response_lid |
            response_xy |
            response_str |
            response_num |
            response_grp |
            response_extension)+
            )>
    <!ATTLIST flow  class CDATA  'Block' >"""
    XMLNAME = 'flow'
    XMLATTR_class = 'flow_class'
    XMLCONTENT = xml.ElementContent

    def __init__(self, parent):
        FlowContainer.__init__(self, parent)

    def is_inline(self):
        """flow is always treated as a block if flow_class is specified,
        otherwise it is treated as a block unless it is an only child."""
        if not self.flow_class:
            return self.inline_children()
        else:
            return False

    def migrate_content_to_v2(self, parent, child_type, log):
        """flow typically maps to a div element.

        A flow with a specified class always becomes a div
        A flow with inline children generates a paragraph to hold them
        A flow with no class is ignored."""
        if self.flow_class:
            if child_type in (html.BlockMixin, html.FlowMixin):
                div = parent.add_child(
                    html.Div, (qtiv2.core.IMSQTI_NAMESPACE, 'div'))
                div.style_class = self.flow_class
                FlowContainer.migrate_content_to_v2(self, div, html.FlowMixin,
                                                    log)
            else:
                span = parent.add_child(
                    html.Span, (qtiv2.core.IMSQTI_NAMESPACE, 'span'))
                span.style_class = self.flow_class
                FlowContainer.migrate_content_to_v2(
                    self, span, html.InlineMixin, log)
        else:
            FlowContainer.migrate_content_to_v2(self, parent, child_type, log)


class Response(common.ContentMixin, core.QTIElement):

    """Abstract class to act as a parent for all response_* elements."""

    def __init__(self, parent):
        core.QTIElement.__init__(self, parent)
        common.ContentMixin.__init__(self)


class ResponseExtension(Response):

    """This element supports proprietary alternatives to the current range of
    'response' elements::

    <!ELEMENT response_extension ANY>"""
    XMLNAME = "response_extension"
    XMLCONTENT = xml.XMLMixedContent


class ResponseThing(Response):

    """Abstract class for the main response_* elements::

    <!ELEMENT response_* ((material | material_ref)? ,
            (render_choice | render_hotspot | render_slider | render_fib |
            render_extension) ,
            (material | material_ref)?)>
    <!ATTLIST response_*
            rcardinality	(Single | Multiple | Ordered )  'Single'
    rtiming  		(Yes | No )  'No'
            ident 			CDATA  #REQUIRED >"""
    XMLATTR_ident = 'ident'
    XMLATTR_rcardinality = (
        'rCardinality',
        core.RCardinality.from_str_title,
        core.RCardinality.to_str)
    XMLATTR_rtiming = ('rTiming', core.ParseYesNo, core.FormatYesNo)
    XMLATTR_ident = 'ident'
    XMLCONTENT = xml.ElementContent

    def __init__(self, parent):
        Response.__init__(self, parent)
        self.ident = None
        self.rCardinality = core.RCardinality.DEFAULT
        self.rTiming = False
        self.intro = []
        self.prompt = []
        self.inlinePrompt = True
        self.render = None
        self.outro = []
        self.footer = []
        self.inlineFooter = True

    def add_child(self, child_class, name=None):
        if issubclass(child_class, (common.Material, common.MaterialRef)):
            child = child_class(self)
            if self.render:
                self.outro.append(child)
            else:
                self.intro.append(child)
            if name:
                child.set_xmlname(name)
            return child
        elif issubclass(child_class, Render):
            child = child_class(self)
            self.render = child
            if name:
                child.set_xmlname(name)
            return child
        elif issubclass(child_class, common.ContentMixin):
            raise TypeError
        else:
            return super(ResponseThing, self).add_child(child_class, name)

    def get_children(self):
        for child in self.intro:
            yield child
        if self.render:
            yield self.render
        for child in self.outro:
            yield child

    def content_changed(self):
        if isinstance(self.render, RenderFIB) and self.render.mixed_model():
            # use simplified prompt logic.
            self.prompt = self.intro
            self.inlintPrompt = True
            for child in self.prompt:
                if not child.is_inline():
                    self.inlinePrompt = False
            self.footer = self.outro
            self.inlineFooter = True
            for child in self.footer:
                if not child.is_inline():
                    self.inlineFooter = False
        elif self.render:
            # all the material up to the first response_label is the prompt
            self.prompt = []
            self.inlinePrompt = True
            render_children = self.render.get_label_content()
            for child in self.intro + render_children:
                # print child.__class__,child.xmlname
                if isinstance(child, ResponseLabel):
                    break
                self.prompt.append(child)
                if not child.is_inline():
                    self.inlinePrompt = False
            self.footer = []
            self.inlineFooter = True
            found_label = False
            for child in render_children + self.outro:
                if isinstance(child, ResponseLabel):
                    self.footer = []
                    self.inlineFooter = True
                    found_label = True
                    continue
                if found_label:
                    self.footer.append(child)
                    if not child.is_inline():
                        self.inlineFooter = False

    def inline_children(self):
        return self.inlinePrompt and (
            self.render is None or
            self.render.is_inline()) and self.inlineFooter

    def get_base_type(self, interaction):
        """Returns the base type to use for the given interaction."""
        raise core.QTIUnimplementedError(
            "BaseType selection for %s" % self.__class__.__name__)

    def migrate_content_to_v2(self, parent, child_type, log):
        if self.inlinePrompt:
            interaction_prompt = self.prompt
        else:
            if child_type is html.InlineMixin:
                raise core.QTIError("Unexpected attempt to inline interaction")
            interaction_prompt = None
            if isinstance(self.render, RenderHotspot):
                div = parent.add_child(
                    html.Div, (qtiv2.core.IMSQTI_NAMESPACE, 'div'))
                common.ContentMixin.migrate_content_to_v2(
                    self, div, html.FlowMixin, log, self.prompt)
                # Now we need to find any images and pass them to render
                # hotspot instead which we do by reusing the interaction_prompt
                # (currently we only find the first image).
                interaction_prompt = list(
                    div.find_children_depth_first(html.Img, False))[0:1]
            else:
                common.ContentMixin.migrate_content_to_v2(
                    self, parent, child_type, log, self.prompt)
        if self.render:
            interaction_list = self.render.migrate_interaction_to_v2(
                parent, child_type, interaction_prompt, log)
            item = parent.find_parent(qtiv2.items.AssessmentItem)
            if (len(interaction_list) > 1 and
                    self.rCardinality == core.RCardinality.Single):
                log.append(
                    "Error: unable to migrate a response with Single"
                    " cardinality to a single interaction: %s" %
                    self.ident)
                interaction_list = []
                response_list = []
            else:
                base_identifier = qtiv2.core.ValidateIdentifier(self.ident)
                response_list = []
                if len(interaction_list) > 1:
                    i = 0
                    for interaction in interaction_list:
                        i = i + 1
                        while True:
                            ridentifier = "%s_%02i" % (base_identifier, i)
                            if (item is None or
                                    not item.IsDeclared(ridentifier)):
                                break
                        interaction.responseIdentifier = ridentifier
                        response_list.append(ridentifier)
                elif interaction_list:
                    interaction = interaction_list[0]
                    interaction.responseIdentifier = base_identifier
                    response_list = [interaction.responseIdentifier]
            if item:
                for i, r in zip(interaction_list, response_list):
                    d = item.add_child(qtiv2.variables.ResponseDeclaration)
                    d.identifier = r
                    d.cardinality = core.MigrateV2Cardinality(
                        self.rCardinality)
                    d.baseType = self.get_base_type(interaction_list[0])
                    self.render.migrate_interaction_default_to_v2(d, i)
                    item.RegisterDeclaration(d)
                if len(response_list) > 1:
                    d = item.add_child(qtiv2.variables.OutcomeDeclaration)
                    d.identifier = base_identifier
                    d.cardinality = core.MigrateV2Cardinality(
                        self.rCardinality)
                    d.baseType = self.get_base_type(interaction_list[0])
                    item.RegisterDeclaration(d)
                    # now we need to fix this outcome up with a value in
                    # response processing
                    self_item = self.find_parent(Item)
                    if self_item:
                        for rp in self_item.ResProcessing:
                            rp._interactionFixup[base_identifier] = \
                                response_list
        # the footer is in no-man's land so we just back-fill
        common.ContentMixin.migrate_content_to_v2(
            self, parent, child_type, log, self.footer)


class ResponseLId(ResponseThing):

    """The <response_lid> element contains the instructions for the
    presentation of questions whose response will be the logical label
    of the selected answer. The question can be rendered in a variety of
    ways depending on the way in which the material is to be presented
    to the participant::

    <!ELEMENT response_lid ((material | material_ref)? ,
            (render_choice | render_hotspot | render_slider | render_fib |
            render_extension) ,
            (material | material_ref)?)>
    <!ATTLIST response_lid
            rcardinality	(Single | Multiple | Ordered )  'Single'
            rtiming  		(Yes | No )  'No'
            ident 			CDATA  #REQUIRED >"""
    XMLNAME = 'response_lid'

    def get_base_type(self, interaction):
        """We always return identifier for response_lid."""
        return qtiv2.variables.BaseType.identifier


class ResponseXY(ResponseThing):

    """The <response_xy> element contains the instructions for the
    presentation of questions whose response will be the 'x-y'
    co-ordinate of the selected answer. The question can be rendered in
    a variety of ways depending on the way in which the material is to
    be presented to the participant::

    <!ELEMENT response_xy ((material | material_ref)? ,
            (render_choice | render_hotspot | render_slider | render_fib |
            render_extension) ,
            (material | material_ref)?)>
    <!ATTLIST response_xy
            rcardinality	(Single | Multiple | Ordered )  'Single'
            rtiming  		(Yes | No )  'No'
            ident 			CDATA  #REQUIRED >
    """
    XMLNAME = 'response_xy'

    def get_base_type(self, interaction):
        """For select point we return a point for response_xy."""
        if isinstance(interaction, qtiv2.interactions.SelectPointInteraction):
            return qtiv2.variables.BaseType.point
        else:
            return ResponseThing.get_base_type(self)


class ResponseStr(ResponseThing):

    """The <response_str> element contains the instructions for the
    presentation of questions whose response will be the a string. The
    question can be rendered in a variety of ways depending on the way
    in which the material is to be presented to the participant::

    <!ELEMENT response_str ((material | material_ref)? ,
            (render_choice | render_hotspot | render_slider | render_fib |
            render_extension) ,
            (material | material_ref)?)>

    <!ATTLIST response_str
            rcardinality	(Single | Multiple | Ordered )  'Single'
            ident 			CDATA #REQUIRED
            rtiming  		(Yes | No )  'No' >"""
    XMLNAME = 'response_str'

    def get_base_type(self, interaction):
        """We always return string for response_str."""
        return qtiv2.variables.BaseType.string


class ResponseNum(ResponseThing):

    """The <response_num> element contains the instructions for the
    presentation of questions whose response will be a number. The
    question can be rendered in a variety of ways depending on the way
    in which the material is to be presented to the participant::

    <!ELEMENT response_num ((material | material_ref)? ,
            (render_choice | render_hotspot | render_slider | render_fib |
            render_extension) ,
            (material | material_ref)?)>
    <!ATTLIST response_num
            numtype			(Integer | Decimal | Scientific )  'Integer'
            rcardinality	(Single | Multiple | Ordered )  'Single'
            ident CDATA  	#REQUIRED
            rtiming  		(Yes | No )  'No' >
    """
    XMLNAME = 'response_num'
    XMLATTR_numtype = (
        'numType', core.NumType.from_str_title, core.NumType.to_str)
    XMLCONTENT = xml.ElementContent

    def __init__(self, parent):
        ResponseThing.__init__(self, parent)
        self.numType = core.NumType.Integer

    def get_base_type(self, interaction):
        """Returns integer for numtype of "Integer", float otherwise."""
        if self.numType == core.NumType.Integer:
            return qtiv2.variables.BaseType.integer
        else:
            return qtiv2.variables.BaseType.float


class ResponseGrp(common.ContentMixin, core.QTIElement):

    """The <response_grp> element contains the instructions for the
    presentation of questions whose response will be a group of logical
    identifiers. The question can be rendered in a variety of ways
    depending on the way in which the material is to be presented to the
    participant::

    <!ELEMENT response_grp ((material | material_ref)? ,
            (render_choice | render_hotspot | render_slider | render_fib |
            render_extension) ,
            (material | material_ref)?)>
    <!ATTLIST response_grp
            rcardinality	(Single | Multiple | Ordered )  'Single'
            ident			CDATA  #REQUIRED
            rtiming			(Yes | No )  'No' >"""
    XMLNAME = 'response_grp'
    XMLCONTENT = xml.ElementContent


class Render(common.ContentMixin, core.QTIElement):

    """Abstract class to act as a parent for all render_* elements."""

    def __init__(self, parent):
        core.QTIElement.__init__(self, parent)
        common.ContentMixin.__init__(self)


class RenderThing(Render):

    """Abstract base class for all render_* objects::

    <!ELEMENT render_* ((material | material_ref | response_label |
    flow_label)* , response_na?)>"""
    XMLCONTENT = xml.ElementContent

    def __init__(self, parent):
        Render.__init__(self, parent)
        self.ResponseNA = None

    def content_child(self, child_class):
        return child_class in (common.Material, common.MaterialRef,
                               ResponseLabel, FlowLabel)

    def get_children(self):
        for child in self.get_content_children():
            yield child
        if self.ResponseNA:
            yield self.ResponseNA

    def get_label_content(self):
        """Returns a flat list of content items, stripping away flow_labels."""
        children = []
        for child in self.get_content_children():
            if isinstance(child, FlowLabel):
                children = children + child.get_label_content()
            else:
                children.append(child)
        return children

    def is_inline(self):
        for child in self.get_label_content():
            if not child.is_inline():
                return False
        return True

    def migrate_interaction_to_v2(self, parent, child_type, prompt, log):
        raise core.QTIUnimplementedError(
            "%s x %s" %
            (self.parent.__class__.__name__, self.__class__.__name__))

    def migrate_interaction_default_to_v2(self, parent, interaction):
        # Most interactions do not need default values.
        pass


class RenderChoice(RenderThing):

    """The <render_choice> element instructs the question-engine to render the
    question using a classical multiple-choice format. The number of possible
    responses is determined by the <response_label> elements contained. Both
    flowed and non-flowed formats are supported::

    <!ELEMENT render_choice ((material | material_ref | response_label |
    flow_label)* , response_na?)>
    <!ATTLIST render_choice
            shuffle     (Yes | No )  'No'
            minnumber	CDATA  #IMPLIED
            maxnumber	CDATA  #IMPLIED >"""
    XMLNAME = 'render_choice'
    XMLATTR_shuffle = ('shuffle', core.ParseYesNo, core.FormatYesNo)
    XMLATTR_minnumber = ('minNumber', xsi.integer_from_str, xsi.integer_to_str)
    XMLATTR_maxnumber = ('maxNumber', xsi.integer_from_str, xsi.integer_to_str)
    XMLCONTENT = xml.ElementContent

    def __init__(self, parent):
        RenderThing.__init__(self, parent)
        self.shuffle = False
        self.minNumber = None
        self.maxNumber = None

    def is_inline(self):
        """This always results in a block-like interaction."""
        return False

    def migrate_interaction_to_v2(self, parent, child_type, prompt, log):
        """Migrates this content to v2 adding it to the parent content node."""
        interaction = None
        if isinstance(self.parent, ResponseLId):
            if child_type is html.InlineMixin:
                raise core.QTIError(
                    "Unexpected attempt to put block interaction in"
                    " inline context")
            if self.parent.rCardinality == core.RCardinality.Ordered:
                raise core.QTIUnimplementedError("OrderInteraction")
            else:
                interaction = parent.add_child(
                    qtiv2.interactions.ChoiceInteraction)
        else:
            raise core.QTIUnimplementedError(
                "%s x render_choice" % self.parent.__class__.__name__)
        if prompt:
            interaction_prompt = interaction.add_child(
                qtiv2.interactions.Prompt)
            for child in prompt:
                child.migrate_content_to_v2(
                    interaction_prompt, html.InlineMixin, log)
        if self.minNumber is not None:
            interaction.minChoices = self.minNumber
        if self.maxNumber is not None:
            interaction.maxChoices = self.maxNumber
        interaction.shuffle = self.shuffle
        for child in self.get_label_content():
            if isinstance(child, ResponseLabel):
                child.migrate_simple_choice_to_v2(interaction, log)
        return [interaction]


class RenderHotspot(RenderThing):

    """The <render_hotspot> element instructs the question-engine to render the
    question using a classical image hot-spot format. The number of possible
    responses is determined by the <response_label> elements contained. Both
    flowed and non-flowed formats are supported::

    <!ELEMENT render_hotspot ((material | material_ref | response_label |
    flow_label)* , response_na?)>
    <!ATTLIST render_hotspot
            maxnumber	CDATA  #IMPLIED
            minnumber	CDATA  #IMPLIED
            showdraw    (Yes | No )  'No' >"""
    XMLNAME = 'render_hotspot'
    XMLATTR_maxnumber = ('maxNumber', xsi.integer_from_str, xsi.integer_to_str)
    XMLATTR_minnumber = ('minNumber', xsi.integer_from_str, xsi.integer_to_str)
    XMLATTR_showdraw = ('showDraw', core.ParseYesNo, core.FormatYesNo)
    XMLCONTENT = xml.ElementContent

    def __init__(self, parent):
        RenderThing.__init__(self, parent)
        self.maxNumber = None
        self.minNumber = None
        self.showDraw = False

    def is_inline(self):
        """This always results in a block-like interaction."""
        return False

    def migrate_interaction_to_v2(self, parent, child_type, prompt, log):
        """Migrates this content to v2 adding it to the parent content node."""
        interaction = None
        if isinstance(self.parent, ResponseLId):
            if self.parent.rCardinality == core.RCardinality.Ordered:
                raise core.QTIUnimplementedError("GraphicOrderInteraction")
            else:
                interaction = parent.add_child(
                    qtiv2.interactions.HotspotInteraction)
        elif isinstance(self.parent, ResponseXY):
            if self.parent.rCardinality == core.RCardinality.Ordered:
                raise core.QTIUnimplementedError(
                    'response_xy x render_hotspot')
            else:
                interaction = parent.add_child(
                    qtiv2.interactions.SelectPointInteraction)
        else:
            raise core.QTIUnimplementedError(
                "%s x render_hotspot" % self.parent.__class__.__name__)
        if self.showDraw:
            log.append(
                'Warning: ignoring showdraw="Yes", what did you really want'
                ' to happen?')
        if self.minNumber is not None:
            interaction.minChoices = self.minNumber
        if self.maxNumber is not None:
            interaction.maxChoices = self.maxNumber
        labels = list(self.find_children_depth_first(ResponseLabel, False))
        # prompt is either a single <img> tag we already migrated or.. a
        # set of inline objects that are still to be migrated (and which
        # should contain the hotspot image).
        img = None
        interaction_prompt = None
        hotspot_image = interaction.add_child(
            html.Object, (qtiv2.core.IMSQTI_NAMESPACE, 'object'))
        if prompt:
            if not isinstance(prompt[0], html.Img):
                interaction_prompt = interaction.add_child(
                    qtiv2.interactions.Prompt)
                for child in prompt:
                    child.migrate_content_to_v2(
                        interaction_prompt, html.InlineMixin, log)
                prompt = list(
                    interaction_prompt.find_children_depth_first(
                        html.Img,
                        False))[
                    0:1]
        if prompt:
            # now the prompt should be a list containing a single image to use
            # as the hotspot
            img = prompt[0]
            hotspot_image.data = img.src
            hotspot_image.height = img.height
            hotspot_image.width = img.width
            if img.src:
                # Annoyingly, Img throws away mime-type information from
                # matimage
                images = list(
                    self.parent.find_children_depth_first(
                        common.MatImage,
                        False))[
                    0:1]
                if images and images[0].uri:
                    # Check that this is the right image in case img was
                    # embedded in MatText
                    if (str(images[0].resolve_uri(images[0].uri)) ==
                            str(img.resolve_uri(img.src))):
                        hotspot_image.type = images[0].imageType
            for child in labels:
                if isinstance(child, ResponseLabel):
                    child.migrate_hotspot_choice_to_v2(interaction, log)
            return [interaction]
        else:
            # tricky, let's start by getting all the matimage elements in the
            # presentation
            images = []
            presentation = self.find_parent(Presentation)
            if presentation:
                images = list(
                    presentation.find_children_depth_first(
                        common.MatImage,
                        False))
            hsi = []
            if len(images) == 1:
                # Single image that must have gone AWOL
                hsi.append((images[0], labels))
            else:
                # multiple images are scanned for those at fixed positions in
                # the presentation which are hit by a hotspot (interpreted
                # relative to the presentation).
                for img in images:
                    hits = []
                    for child in labels:
                        if child.hotspot_in_image(img):
                            hits.append(child)
                    if hits:
                        # So some of our hotspots hit this image
                        hsi.append((img, hits))
            if len(hsi) == 0:
                log.append(
                    "Error: omitting render_hotspot with no hotspot image")
                return []
            else:
                img, hits = hsi[0]
                hotspot_image.data = img.resolve_uri(img.uri)
                hotspot_image.type = img.imageType
                hotspot_image.height = html.LengthType(img.height)
                hotspot_image.width = html.LengthType(img.width)
                # it will get cleaned up later
                if len(hsi) > 0:
                    # Worst case: multiple images => multiple hotspot
                    # interactions
                    if self.maxNumber is not None:
                        log.append(
                            "Warning: multi-image hotspot maps to multiple"
                            " interactions, maxChoices can no longer be"
                            " enforced")
                    interaction.minChoices = None
                    for child in hits:
                        child.migrate_hotspot_choice_to_v2(
                            interaction, log, img.x0, img.y0)
                    interaction_list = [interaction]
                    for img, hits in hsi[1:]:
                        interaction = parent.add_child(
                            qtiv2.interactions.HotspotInteraction)
                        if self.maxNumber is not None:
                            interaction.maxChoices = self.maxNumber
                        hotspot_image = interaction.add_child(
                            html.Object, (qtiv2.core.IMSQTI_NAMESPACE,
                                          'object'))
                        hotspot_image.data = img.resolve_uri(img.uri)
                        hotspot_image.type = img.imageType
                        hotspot_image.height = html.LengthType(img.height)
                        hotspot_image.width = html.LengthType(img.width)
                        for child in hits:
                            child.migrate_hotspot_choice_to_v2(
                                interaction, log, img.x0, img.y0)
                        interaction_list.append(interaction)
                    return interaction_list
                else:
                    # Best case: single image that just went AWOL
                    for child in labels:
                        child.migrate_hotspot_choice_to_v2(
                            interaction, log, img.x0, img.y0)
                    return [interaction]


class RenderFIB(RenderThing):

    """The <render_fib> element instructs the question-engine to render the
    question using a classical fill-in-blank format. The number of possible
    responses is determined by the <response_label> elements contained. Both
    flowed and non-flowed formats are supported::

    <!ELEMENT render_fib ((material | material_ref | response_label |
    flow_label)* , response_na?)>
    <!ATTLIST render_fib
            encoding	CDATA  'UTF_8'
            fibtype		(String | Integer | Decimal | Scientific )  'String'
            rows		CDATA  #IMPLIED
            maxchars	CDATA  #IMPLIED
            prompt		(Box | Dashline | Asterisk | Underline )  #IMPLIED
            columns		CDATA  #IMPLIED
            charset		CDATA  'ascii-us'
            maxnumber	CDATA  #IMPLIED
            minnumber	CDATA  #IMPLIED >"""
    XMLNAME = 'render_fib'
    XMLATTR_encoding = 'encoding'
    XMLATTR_fibtype = (
        'fibType', core.FIBType.from_str_title, core.FIBType.to_str)
    XMLATTR_rows = ('rows', xsi.integer_from_str, xsi.integer_to_str)
    XMLATTR_maxchars = ('maxChars', xsi.integer_from_str, xsi.integer_to_str)
    XMLATTR_prompt = (
        'prompt',
        core.PromptType.from_str_title,
        core.PromptType.to_str)
    XMLATTR_columns = ('columns', xsi.integer_from_str, xsi.integer_to_str)
    XMLATTR_charset = 'charset'
    XMLATTR_maxnumber = ('maxNumber', xsi.integer_from_str, xsi.integer_to_str)
    XMLATTR_minnumber = ('minNumber', xsi.integer_from_str, xsi.integer_to_str)
    XMLCONTENT = xml.ElementContent

    def __init__(self, parent):
        RenderThing.__init__(self, parent)
        self.encoding = 'UTF_8'
        self.fibType = core.FIBType.String
        self.rows = None
        self.maxChars = None
        self.prompt = None
        self.columns = None
        self.charset = 'ascii-us'
        self.minNumber = None
        self.maxNumber = None
        self.labels = []

    def content_changed(self):
        self.labels = list(
            self.find_children_depth_first(ResponseLabel, False))

    def mixed_model(self):
        """Indicates whether or not this FIB uses a mixed model or not.

        A mixed model means that render_fib is treated as a mixture of
        interaction and content elements.  In an unmixed model the render_fib
        is treated as a single block interaction with an optional prompt.

        If the render_fib contains content, followed by labels, then we treat
        it as a prompt + fib and return False

        If the render_fib contains a mixture of content and labels, then we
        return True

        If the render_fib contains no content at all we assume it needs to be
        mixed into the surrounding content and return True."""
        children = self.get_label_content()
        found_label = False
        found_content = False
        for child in children:
            if isinstance(child, ResponseLabel):
                found_label = True
            elif found_label:
                # any content after the first label means mixed mode.
                return True
            else:
                found_content = True
        return not found_content

    def is_inline(self):
        if self.mixed_model():
            return RenderThing.is_inline(self)
        else:
            return False

    def inline_fib_label(self):
        if self.rows is None or self.rows == 1:
            return True
        else:
            return False

    def migrate_fib_label_to_v2(self, label, parent, child_type, log):
        if self.inline_fib_label() or child_type is html.InlineMixin:
            parent.add_child(qtiv2.interactions.TextEntryInteraction)
        else:
            parent.add_child(qtiv2.interactions.ExtendedTextInteraction)
        if list(label.get_children()):
            log.append(
                "Warning: ignoring content in render_fib.response_label")

    def migrate_interaction_to_v2(self, parent, child_type, prompt, log):
        if self.inline_fib_label() or child_type is html.InlineMixin:
            interaction_type = qtiv2.interactions.TextEntryInteraction
        else:
            interaction_type = qtiv2.interactions.ExtendedTextInteraction
        interaction_list = list(
            parent.find_children_depth_first(interaction_type, False))
        icount = len(interaction_list)
        # now migrate this object
        RenderThing.migrate_content_to_v2(self, parent, child_type, log)
        interaction_list = list(
            parent.find_children_depth_first(interaction_type, False))
        # ignore any pre-existing interactions of this type
        interaction_list = interaction_list[icount:]
        if (self.parent.rCardinality == core.RCardinality.Single and
                len(interaction_list) > 1):
            log.append(
                "Warning: single response fib ignoring all but last"
                " <response_label>")
            for interaction in interaction_list[:-1]:
                interaction.parent.remove_child(interaction)
            interaction_list = interaction_list[-1:]
        for interaction in interaction_list:
            if self.maxChars is not None:
                interaction.expectedLength = self.maxChars
            elif self.rows is not None and self.columns is not None:
                interaction.expectedLength = self.rows * self.columns
            if interaction_type is qtiv2.interactions.ExtendedTextInteraction:
                if self.rows is not None:
                    interaction.expectedLines = self.rows
        return interaction_list


class RenderSlider(RenderThing):

    """The <render_slider> element instructs the question-engine to render the
    question using dynamic slider. The number of possible responses is
    determined by the <response_label> elements contained. Both flowed and
    non-flowed formats are supported::

    <!ELEMENT render_slider ((material | material_ref | response_label |
    flow_label)* , response_na?)>
    <!ATTLIST render_slider
            orientation		(Horizontal | Vertical )  'Horizontal'
            lowerbound		CDATA  #REQUIRED
            upperbound		CDATA  #REQUIRED
            step			CDATA  #IMPLIED
            startval		CDATA  #IMPLIED
            steplabel		(Yes | No )  'No'
            maxnumber		CDATA  #IMPLIED
            minnumber		CDATA  #IMPLIED >"""
    XMLNAME = 'render_slider'
    XMLATTR_orientation = (
        'orientation',
        core.Orientation.from_str_title,
        core.Orientation.to_str)
    XMLATTR_lowerbound = (
        'lowerBound', xsi.integer_from_str, xsi.integer_to_str)
    XMLATTR_upperbound = (
        'upperBound', xsi.integer_from_str, xsi.integer_to_str)
    XMLATTR_step = ('step', xsi.integer_from_str, xsi.integer_to_str)
    XMLATTR_startval = ('startVal', xsi.integer_from_str, xsi.integer_to_str)
    XMLATTR_steplabel = ('stepLabel', core.ParseYesNo, core.FormatYesNo)
    XMLATTR_maxnumber = ('maxNumber', xsi.integer_from_str, xsi.integer_to_str)
    XMLATTR_minnumber = ('minNumber', xsi.integer_from_str, xsi.integer_to_str)
    XMLCONTENT = xml.ElementContent

    def __init__(self, parent):
        RenderThing.__init__(self, parent)
        self.orientation = core.Orientation.Horizontal
        self.lowerBound = None
        self.upperBound = None
        self.step = None
        self.startVal = None
        self.stepLabel = False
        self.minNumber = None
        self.maxNumber = None

    def is_inline(self):
        """This always results in a block-like interaction."""
        return False

    def migrate_interaction_to_v2(self, parent, child_type, prompt, log):
        """Migrates this content to v2 adding it to the parent content node."""
        interaction = None
        labels = list(self.find_children_depth_first(ResponseLabel, False))
        if self.maxNumber is None:
            max_choices = len(labels)
        else:
            max_choices = self.maxNumber
        if isinstance(self.parent, ResponseLId):
            if self.parent.rCardinality == core.RCardinality.Single:
                log.append(
                    "Warning: choice-slider replaced with"
                    " choiceInteraction.slider")
                interaction = parent.add_child(
                    qtiv2.interactions.ChoiceInteraction)
                interaction.style_class = ['slider']
                interaction.minChoices = 1
                interaction.maxChoices = 1
            elif self.parent.rCardinality == core.RCardinality.Ordered:
                log.append(
                    "Error: ordered-slider replaced with"
                    " orderInteraction.slider")
                raise core.QTIUnimplementedError("OrderInteraction")
            else:
                log.append(
                    "Error: multiple-slider replaced with"
                    " choiceInteraction.slider")
                interaction = parent.add_child(
                    qtiv2.interactions.ChoiceInteraction)
                interaction.style_class = ['slider']
                if self.minNumber is not None:
                    interaction.minChoices = self.minNumber
                else:
                    interaction.minChoices = max_choices
                interaction.maxChoices = max_choices
            interaction.shuffle = False
            for child in labels:
                child.migrate_simple_choice_to_v2(interaction, log)
        elif isinstance(self.parent, ResponseNum):
            if self.parent.rCardinality == core.RCardinality.Single:
                interaction = parent.add_child(
                    qtiv2.interactions.SliderInteraction)
                interaction.lowerBound = float(self.lowerBound)
                interaction.upperBound = float(self.upperBound)
                if self.step is not None:
                    interaction.step = self.step
                if self.orientation is not None:
                    interaction.orientation = core.MigrateV2Orientation(
                        self.orientation)
                # startValues are handled below after the variable is declared
            else:
                raise core.QTIUnimplementedError(
                    "Multiple/Ordered SliderInteraction")
        else:
            raise core.QTIUnimplementedError(
                "%s x render_slider" % self.parent.__class__.__name__)
        if prompt:
            interaction_prompt = interaction.add_child(
                qtiv2.interactions.Prompt)
            for child in prompt:
                child.migrate_content_to_v2(
                    interaction_prompt, html.InlineMixin, log)
        return [interaction]

    def migrate_interaction_default_to_v2(self, declaration, interaction):
        # Most interactions do not need default values.
        if (isinstance(interaction, qtiv2.interactions.SliderInteraction) and
                self.startVal is not None):
            value = declaration.add_child(
                qtiv2.variables.DefaultValue).add_child(
                qtiv2.variables.ValueElement)
            if declaration.baseType == qtiv2.variables.BaseType.integer:
                value.set_value(xsi.integer_to_str(self.startVal))
            elif declaration.baseType == qtiv2.variables.BaseType.float:
                value.set_value(xsi.float_to_str(self.startVal))
            else:
                # slider bound to something else?
                raise core.QTIError(
                    "Unexpected slider type for default: %s" %
                    qtiv2.variables.BaseType.to_str(
                        declaration.baseType))


class RenderExtension(Render):

    """This element supports proprietary alternatives to the current range of
    'render' elements::

    <!ELEMENT render_extension ANY>"""
    XMLNAME = "render_extension"
    XMLCONTENT = xml.XMLMixedContent


class ResponseLabel(common.ContentMixin, core.QTIElement):

    """The <response_label> is used to define the possible response
    choices that are presented to the user. This information includes
    the material to be shown to the user and the logical label that is
    associated with that response. The label is used in the response
    processing. Flow and non-flow approaches are supported::

    <!ELEMENT response_label (#PCDATA | qticomment | material | material_ref |
    flow_mat)*>
    <!ATTLIST response_label  rshuffle     (Yes | No )  'Yes'
    rarea			(Ellipse | Rectangle | Bounded )  'Ellipse'
    rrange			(Exact | Range )  'Exact'
    labelrefid		CDATA  #IMPLIED
    ident CDATA		#REQUIRED
    match_group		CDATA  #IMPLIED
    match_max		CDATA  #IMPLIED >"""
    XMLNAME = 'response_label'
    XMLATTR_ident = 'ident'
    XMLATTR_rshuffle = ('rShuffle', core.ParseYesNo, core.FormatYesNo)
    XMLATTR_rarea = (
        'rArea', core.Area.from_str_title, core.Area.to_str)
    XMLCONTENT = xml.XMLMixedContent

    def __init__(self, parent):
        core.QTIElement.__init__(self, parent)
        common.ContentMixin.__init__(self)
        self.ident = ''
        self.rShuffle = True
        self.rArea = None

    def add_child(self, child_class, name=None):
        """Although we inherit from the ContentMixin class we don't
        allow it to capture content-children because this element has
        mixed content - though in practice it really should have either
        data or element content."""
        if issubclass(child_class, common.ContentMixin):
            # add it to the default content model
            child = child_class(self)
            self._children.append(child)
            if name:
                child.set_xmlname(name)
            return child
        else:
            return super(ResponseLabel, self).add_child(child_class, name)

    def is_inline(self):
        """Whether or not a response_label is inline depends on the context...

        render_choice: a choice is a block
        render_hotspot: at most a label on the image so treated as inline
        render_fib: always inline
        """
        render = self.find_parent(RenderThing)
        if isinstance(render, RenderChoice):
            return False
        elif isinstance(render, RenderHotspot):
            return True
        elif isinstance(render, RenderFIB):
            return render.inline_fib_label()
        else:
            return self.inline_children()

    def inline_children(self):
        for child in core.QTIElement.get_children(self):
            if is_text(child):
                continue
            elif issubclass(child.__class__, common.ContentMixin):
                if child.is_inline():
                    continue
                return False
            else:
                # QTIComment most likely
                continue
        return True

    def migrate_content_to_v2(self, parent, child_type, log):
        """Migrates this content to v2 adding it to the parent content node."""
        render = self.find_parent(RenderThing)
        if isinstance(render, RenderFIB):
            render.migrate_fib_label_to_v2(self, parent, child_type, log)

    def migrate_simple_choice_to_v2(self, interaction, log):
        """Migrate this label into a v2 simpleChoice in interaction."""
        choice = interaction.add_child(qtiv2.interactions.SimpleChoice)
        choice.identifier = qtiv2.core.ValidateIdentifier(self.ident)
        if (isinstance(interaction, qtiv2.interactions.ChoiceInteraction) and
                interaction.shuffle):
            choice.fixed = not self.rShuffle
        data = []
        got_elements = False
        for child in core.QTIElement.get_children(self):
            if is_text(child):
                if len(child.strip()):
                    data.append(child)
            elif isinstance(child, common.QTIComment):
                continue
            else:
                got_elements = True
        if data and got_elements:
            log.append(
                'Warning: ignoring PCDATA in <response_label>, "%s"' %
                ' '.join(data))
        elif data:
            for d in data:
                choice.add_data(d)
        else:
            content = []
            for child in core.QTIElement.get_children(self):
                if isinstance(child, common.ContentMixin):
                    content.append(child)
            common.ContentMixin.migrate_content_to_v2(
                self, choice, html.FlowMixin, log, content)

    def migrate_hotspot_choice_to_v2(self, interaction, log, xoffset=0,
                                     yoffset=0):
        """Migrate this label into a v2 hotspotChoice in interaction."""
        if isinstance(interaction, qtiv2.interactions.SelectPointInteraction):
            log.append(
                "Warning: ignoring response_label in"
                " selectPointInteraction (%s)" %
                self.ident)
            return
        choice = interaction.add_child(qtiv2.interactions.HotspotChoice)
        choice.identifier = qtiv2.core.ValidateIdentifier(self.ident)
        # Hard to believe I know, but we sift the content of the
        # response_label into string data (which is parsed for
        # coordinates) and elements which have their text extracted for
        # the hotspot label.
        lang, label_data, value_data = self.parse_value()
        choice.shape, choice.coords = core.MigrateV2AreaCoords(
            self.rArea, value_data, log)
        if xoffset or yoffset:
            qtiv2.core.offset_shape(
                choice.shape, choice.coords, xoffset, yoffset)
        if lang is not None:
            choice.set_lang(lang)
        if label_data:
            choice.hotspotLabel = label_data

    def hotspot_in_image(self, mat_image):
        """Tests this hotspot to see if it overlaps with mat_image.

        The coordinates in the response label are interpreted relative to
        a notional 'stage' on which the presentation takes place.  If the
        image does not have X0,Y0 coordinates then it is ignored and
        we return 0.
        """
        if mat_image.x0 is None or mat_image.y0 is None:
            return False
        if mat_image.width is None or mat_image.height is None:
            return False
        lang, label, value = self.parse_value()
        shape, coords = core.MigrateV2AreaCoords(self.rArea, value, [])
        bounds = qtiv2.core.calculate_shape_bounds(shape, coords)
        if bounds[0] > mat_image.x0 + mat_image.width or bounds[2] < \
                mat_image.x0:
            return False
        if (bounds[1] > mat_image.y0 + mat_image.height or
                bounds[3] < mat_image.y0):
            return False
        return True

    def parse_value(self):
        """Returns lang,label,coords parsed from the value."""
        value_data = []
        label_data = []
        lang = None
        for child in self.get_children():
            if is_text(child):
                value_data.append(child)
            else:
                childLang, text = child.extract_text()
                if lang is None and childLang is not None:
                    lang = childLang
                label_data.append(text)
        value_data = ' '.join(value_data)
        label_data = ' '.join(label_data)
        return lang, label_data, value_data


class FlowLabel(
        common.FlowMixin,
        common.ContentMixin,
        common.QTICommentContainer):

    """The <flow_label> element is the blocking/paragraph equivalent to the
    <response_label> element::

    <!ELEMENT flow_label	(qticomment? , (flow_label | response_label)+)>
    <!ATTLIST flow_label	class CDATA  'Block' >"""
    XMLNAME = 'flow_label'
    XMLCONTENT = xml.ElementContent

    def __init__(self, parent):
        common.QTICommentContainer.__init__(self, parent)
        common.ContentMixin.__init__(self)

    def content_child(self, child_class):
        return child_class is FlowLabel or issubclass(
            child_class, ResponseLabel)

    def get_children(self):
        return itertools.chain(
            common.QTICommentContainer.get_children(self),
            self.contentChildren)

    def get_label_content(self):
        children = []
        for child in self.contentChildren:
            if isinstance(child, FlowLabel):
                children = children + child.get_label_content()
            else:
                children.append(child)
        return children


class ResponseNA(core.QTIElement):

    """This element supports proprietary alternatives for defining the response
    to be created when a participant makes no response to an Item::

    <!ELEMENT response_na ANY>"""
    XMLNAME = 'response_na'
    XMLCONTENT = xml.XMLMixedContent


class ResProcessing(common.QTICommentContainer):

    """This is the element within which all of the instructions for the
    response processing are contained. This includes the scoring
    variables to contain the associated scores and the set of response
    condition tests that are to be applied to the received user
    response::

    <!ELEMENT resprocessing (qticomment? , outcomes ,
    (respcondition | itemproc_extension)+)>
    <!ATTLIST resprocessing  scoremodel CDATA  #IMPLIED >"""
    XMLNAME = 'resprocessing'
    XMLATTR_scoremodel = 'scoreModel'
    XMLCONTENT = xml.ElementContent

    def __init__(self, parent):
        common.QTICommentContainer.__init__(self, parent)
        self.scoreModel = None
        self.Outcomes = Outcomes(self)
        self.ConditionMixin = []

    def get_children(self):
        for child in common.QTICommentContainer.get_children(self):
            yield child
        yield self.Outcomes
        for child in self.ConditionMixin:
            yield child

    def migrate_to_v2(self, v2_item, log):
        """Migrates v1 resprocessing to v2 ResponseProcessing."""
        rp = v2_item.add_child(qtiv2.processing.ResponseProcessing)
        for outcomeFixup in sorted(self._interactionFixup.keys()):
            set_value = rp.add_child(qtiv2.processing.SetOutcomeValue)
            set_value.identifier = outcomeFixup
            multi = set_value.add_child(qtiv2.expressions.Multiple)
            for rID in self._interactionFixup[outcomeFixup]:
                var = multi.add_child(qtiv2.expressions.Variable)
                var.identifier = rID
        self.Outcomes.migrate_to_v2(v2_item, log)
        cmode = True
        rule_container = rp
        for condition in self.ConditionMixin:
            cmode, rule_container = condition.migrate_rule_to_v2(
                cmode, rule_container, log)


class Outcomes(common.QTICommentContainer):

    """The <outcomes> element contains all of the variable declarations
    that are to be made available to the scoring algorithm. Each
    variable is declared using the <decvar> element apart from the
    default variable called 'SCORE' that is an integer and has a default
    value of zero (0)::

    <!ELEMENT outcomes (qticomment? , (decvar , interpretvar*)+)>

    The implementation of this element takes a liberty with the content model
    because, despite the formulation above, the link between variables and
    their interpretation is not related to the order of the elements within
    the outcomes element.  (An interpretation without a variable reference
    defaults to an interpretation of the default 'SCORE' outcome.)

    When we output this element we do the decvars first, followed by
    the interpretVars."""
    XMLNAME = 'outcomes'
    XMLCONTENT = xml.ElementContent

    def __init__(self, parent):
        common.QTICommentContainer.__init__(self, parent)
        self.DecVar = []
        self.InterpretVar = []

    def get_children(self):
        return itertools.chain(
            common.QTICommentContainer.get_children(self),
            self.DecVar,
            self.InterpretVar)

    def migrate_to_v2(self, v2_item, log):
        for d in self.DecVar:
            d.migrate_to_v2(v2_item, log)
        for i in self.InterpretVar:
            i.migrate_to_v2(v2_item, log)


class ConditionMixin:

    """Mixin class to identify condition elements::

    (respcondition | itemproc_extension)"""
    pass


class RespCondition(common.QTICommentContainer, ConditionMixin):

    """This element contains the actual test to be applied to the user
    responses to determine their correctness or otherwise. Each
    <respcondition> contains an actual test, the assignment of a value
    to the associate scoring variables and the identification of the
    feedback to be associated with the test::

    <!ELEMENT respcondition (qticomment? , conditionvar , setvar* ,
    displayfeedback* , respcond_extension?)>
    <!ATTLIST respcondition
            continue  (Yes | No )  'No'
            title CDATA  #IMPLIED >"""
    XMLNAME = 'respcondition'
    XMLATTR_continue = ('continueFlag', core.ParseYesNo, core.FormatYesNo)
    XMLATTR_title = 'title'
    XMLCONTENT = xml.ElementContent

    def __init__(self, parent):
        common.QTICommentContainer.__init__(self, parent)
        self.continueFlag = False
        self.title = None
        self.ConditionVar = common.ConditionVar(self)
        self.SetVar = []
        self.DisplayFeedback = []
        self.RespCondExtension = None

    def get_children(self):
        for child in common.QTICommentContainer.get_children(self):
            yield child
        yield self.ConditionVar
        for child in itertools.chain(
                self.SetVar,
                self.DisplayFeedback):
            yield child
        if self.RespCondExtension:
            yield self.RespCondExtension

    def migrate_rule_to_v2(self, cmode, rule_container, log):
        """Converts a response condition into v2 response processing rules.

        This method contains some tricky logic to help implement the confusing
        'continue' attribute of response conditions.  The continue attribute
        is interpreted in the following way:

        True: regardless of whether or not the condition matches, carry on to
        evaluate the next expression.

        False: only evaluate the next expression if the condition fails.

        The incoming cmode tells us if the previous condition set continue mode
        (the default is False on the attribute but the algorithm starts with
        continue mode True as the first rule is always evaluated).

        The way the rules are implemented is best illustrated by example, where
        X(True) represents condition X with continue='Yes' etc::

                R1(True),R2(True|False) becomes...

                if R1.test:
                        R1.rules
                if R2.test:
                        R2.rules

                R1(False),R2(True) becomes...

                if R1.test:
                        R1.rules
                else:
                        if R2.test:
                                R2.rules

                R1(False),R2(False) becomes...

                if R1.test:
                        R1.rules
                elif R2.test:
                        R2.rules"""
        if self.continueFlag:
            if not cmode:
                rule_container = rule_container.add_child(
                    qtiv2.processing.ResponseElse)
            rc = rule_container.add_child(qtiv2.processing.ResponseCondition)
            rc_if = rc.add_child(qtiv2.processing.ResponseIf)
        else:
            if cmode:
                rc = rule_container.add_child(
                    qtiv2.processing.ResponseCondition)
                rule_container = rc
                rc_if = rc.add_child(qtiv2.processing.ResponseIf)
            else:
                rc_if = rule_container.add_child(
                    qtiv2.processing.ResponseElseIf)
        self.ConditionVar.migrate_expression_to_v2(rc_if, log)
        for rule in self.SetVar:
            rule.migrate_rule_to_v2(rc_if, log)
        for rule in self.DisplayFeedback:
            rule.migrate_rule_to_v2(rc_if, log)
        return self.continueFlag, rule_container


class RespCondExtension(core.QTIElement):

    """This element supports proprietary alternatives to the <respcondition>
    element::

    <!ELEMENT respcond_extension ANY>"""
    XMLNAME = "respcond_extension"
    XMLCONTENT = xml.XMLMixedContent


class ItemProcExtension(common.ContentMixin, core.QTIElement, ConditionMixin):

    """This element supports proprietary alternatives to the standard Item
    response processing algorithms. element::

    <!ELEMENT itemproc_extension ANY>"""
    XMLNAME = "itemproc_extension"
    XMLCONTENT = xml.XMLMixedContent

    def __init__(self, parent):
        core.QTIElement.__init__(self, parent)
        common.ContentMixin.__init__(self)

    def migrate_rule_to_v2(self, cmode, rule_container, log):
        """Converts an itemProcExtension into v2 response processing rules.

        We only support one type of extension at the moment, the
        humanrater element used as an illustration in the specification
        examples."""
        for child in self.get_children():
            if is_text(child):
                # ignore data
                continue
            elif child.xmlname == 'humanraterdata':
                # humanraterdata extension, migrate content with appropriate
                # view
                v2_item = rule_container.find_parent(
                    qtiv2.items.AssessmentItem)
                rubric = v2_item.get_or_add_child(
                    qtiv2.content.ItemBody).add_child(
                        qtiv2.content.RubricBlock)
                rubric.view = qtiv2.core.View.scorer
                material = list(
                    child.find_children_depth_first(common.Material, False))
                self.migrate_content_to_v2(rubric, html.BlockMixin, log,
                                           material)
        return cmode, rule_container


class ItemFeedback(common.ContentMixin, core.QTIElement):

    """The container for the feedback that is to be presented as a
    result of the user's responses. The feedback can include hints and
    solutions and both of these can be revealed in a variety of
    different ways::

    <!ELEMENT itemfeedback ((flow_mat | material) | solution | hint)+>
    <!ATTLIST itemfeedback
        view	(All | Administrator | AdminAuthority | Assessor |
                 Author | Candidate | InvigilatorProctor |
                 Psychometrician | Scorer | Tutor ) 'All'
        ident CDATA  #REQUIRED
        title CDATA  #IMPLIED >"""
    XMLNAME = 'itemfeedback'
    XMLATTR_view = ('view', core.View.from_str_lower, core.View.to_str)
    XMLATTR_title = 'title'
    XMLATTR_ident = 'ident'

    XMLCONTENT = xml.ElementContent

    def __init__(self, parent):
        core.QTIElement.__init__(self, parent)
        self.view = core.View.DEFAULT
        common.ContentMixin.__init__(self)
        self.title = None
        self.ident = None

    def content_child(self, child_class):
        return child_class in (common.FlowMat, common.Material, Solution, Hint)

    def get_children(self):
        return itertools.chain(
            core.QTIElement.get_children(self),
            self.contentChildren)

    def migrate_to_v2(self, v2_item, log):
        feedback = v2_item.add_child(qtiv2.items.QTIModalFeedback)
        if self.view not in (core.View.All, core.View.Candidate):
            log.append("Warning: discarding view on feedback (%s)" %
                       core.View.to_str(self.view))
        identifier = qtiv2.core.ValidateIdentifier(self.ident, 'FEEDBACK_')
        feedback.outcomeIdentifier = 'FEEDBACK'
        feedback.showHide = qtiv2.core.ShowHide.show
        feedback.identifier = identifier
        feedback.title = self.title
        common.ContentMixin.migrate_content_to_v2(
            self, feedback, html.FlowMixin, log)


class Solution(common.ContentMixin, common.QTICommentContainer):

    """The <solution> element contains the solution(s) that are to be revealed
    to the participant. When these solutions are revealed is outside the scope
    of the specification::

    <!ELEMENT solution (qticomment? , solutionmaterial+)>
    <!ATTLIST solution  feedbackstyle  (Complete | Incremental | Multilevel |
                                        Proprietary )  'Complete' >"""
    XMLNAME = 'solution'
    XMLATTR_feedbackstyle = (
        'feedbackStyle',
        core.FeedbackStyle.from_str_title,
        core.FeedbackStyle.to_str)
    XMLCONTENT = xml.ElementContent

    def __init__(self, parent):
        common.QTICommentContainer.__init__(self, parent)
        common.ContentMixin.__init__(self)
        self.feedbackStyle = core.FeedbackStyle.DEFAULT

    def content_child(self, child_class):
        return child_class is SolutionMaterial

    def get_children(self):
        return itertools.chain(
            common.QTICommentContainer.get_children(self),
            common.ContentMixin.get_content_children(self))


class FeedbackMaterial(common.ContentMixin, core.QTIElement):

    """Abstract class for solutionmaterial and hintmaterial::

    <!ELEMENT * (material+ | flow_mat+)>"""
    XMLCONTENT = xml.ElementContent

    def __init__(self, parent):
        core.QTIElement.__init__(self, parent)
        common.ContentMixin.__init__(self)

    def content_child(self, child_class):
        return child_class in (common.Material, common.FlowMat)

    def get_children(self):
        return common.ContentMixin.get_content_children(self)


class SolutionMaterial(FeedbackMaterial):

    """This is the container for the materials to be presented to the
    participant as the recommended solution. The solution can be revealed in a
    variety of ways but this information is not contained within the material
    being displayed::

    <!ELEMENT solutionmaterial (material+ | flow_mat+)>"""
    XMLNAME = 'solutionmaterial'


class Hint(common.ContentMixin, common.QTICommentContainer):

    """The <hint> element contains the hint(s) that are to be revealed to the
    participant. When these hints are revealed is outside the scope of the
    specification::

    <!ELEMENT hint (qticomment? , hintmaterial+)>
    <!ATTLIST hint  feedbackstyle  (Complete | Incremental | Multilevel |
                                    Proprietary )  'Complete' >"""
    XMLNAME = 'hint'
    XMLATTR_feedbackstyle = (
        'feedbackStyle',
        core.FeedbackStyle.from_str_title,
        core.FeedbackStyle.to_str)
    XMLCONTENT = xml.ElementContent

    def __init__(self, parent):
        common.QTICommentContainer.__init__(self, parent)
        common.ContentMixin.__init__(self)
        self.feedbackStyle = core.FeedbackStyle.DEFAULT

    def content_child(self, child_class):
        return child_class is HintMaterial

    def get_children(self):
        return itertools.chain(
            common.QTICommentContainer.get_children(self),
            common.ContentMixin.get_content_children(self))


class HintMaterial(FeedbackMaterial):

    """This is the container for the materials to be presented to the
    participant as a hint. The hint can be revealed in a variety of ways but
    this information is not contained within the material being displayed::

    <!ELEMENT hintmaterial (material+ | flow_mat+)>"""
    XMLNAME = 'hintmaterial'



#! /usr/bin/env python

import itertools

from . import common
from . import core
from . import item
from ..xml import structures as xml


class OutcomesProcessing(common.QTICommentContainer):

    """The <outcomes_processing> element is the container for all of the
    outcomes processing instructions for Assessments and Sections. Multiple
    outcomes processing containers can be use when multiple scoring algorithms
    are to be applied to produce the aggregated outcomes. If multiple
    outcomes_processing elements are supplied, it is the intention that all
    should be run and the outcomes from all of them together should be reported
    as the outcomes of the enclosing Section or Assessment. The outcome
    variables defined by each <outcomes_processing> element should be unique
    across all outcomes_processing elements defined by a section. In
    particular, it is an error for multiple <outcomes_processing> elements to
    set the same outcome variable and the results will be undefined::

    <!ELEMENT outcomes_processing (qticomment? , outcomes ,
            objects_condition* , processing_parameter* , map_output* ,
            outcomes_feedback_test*)>
    <!ATTLIST outcomes_processing  scoremodel CDATA  #IMPLIED >"""
    XMLNAME = "outcomes_processing"
    XMLATTR_scoremodel = 'scoreModel'
    XMLCONTENT = xml.ElementContent

    def __init__(self, parent):
        common.QTICommentContainer.__init__(self, parent)
        self.scoreModel = None
        self.Outcomes = item.Outcomes(self)
        self.ObjectsCondition = []
        self.ProcessingParameter = []
        self.MapOutput = []
        self.OutcomesFeedbackTest = []

    def get_children(self):
        for child in common.QTICommentContainer.get_children():
            yield child
        yield self.Outcomes
        for child in itertools.chain(
                self.ObjectsCondition,
                self.ProcessingParameter,
                self.MapOutput,
                self.OutcomesFeedbackTest):
            yield child


class ProcessingParameter(core.QTIElement):

    """This element contains the value of a particular parameter that is to be
    used by the corresponding scoring algorithm. Each parameter has a
    particular meaning to each scoring algorithm i.e. there is no established
    vocabulary for these parameters::

    <!ELEMENT processing_parameter (#PCDATA)>
    <!ATTLIST processing_parameter  pname CDATA  #REQUIRED >"""
    XMLNAME = 'processing_parameter'
    XMLATTR_pname = 'pName'
    XMLCONTENT = xml.XMLMixedContent

    def __init__(self, parent):
        core.QTIElement.__init__(self, parent)
        self.pName = None


class MapOutput(core.QTIElement):

    """This element is used to re-map the named variable to another named
    variable (given in the body of the element). The target variable name must
    have been declared using <decvar> in the <outcomes> element of the
    enclosing <outcomes_processing> element. When a variable is remapped, all
    of its derived variables are to be remapped as well. Thus if remapping
    'SCORE' to 'myScore', 'SCORE.min', 'SCORE.max' and 'SCORE.normalized'
    would be remapped to 'myScore.min', 'myScore.max' and 'myScore.normalized'
    respectively.
    Data-type = string (1-256 chars)::

    <!ELEMENT map_output (#PCDATA)>
    <!ATTLIST map_output  varname CDATA  'SCORE' >"""
    XMLNAME = 'map_output'
    XMLATTR_varname = 'varName'
    XMLCONTENT = xml.XMLMixedContent

    def __init__(self, parent):
        core.QTIElement.__init__(self, parent)
        self.varName = None


class ObjectsCondition(common.QTICommentContainer):

    """Each <objects_condition> element defines a subset of the objects (Items
    and Sections) selected by the selection algorithm that are to be used in
    this score. This allows for the construction of subscores. If no
    <outcomes_metadata> element is present within the <objects_conditions>
    element to select a subset of objects, then the <objects_conditions>
    applies to all objects selected by the selection algorithm. If multiple
    <objects_conditions> are given within a single <outcomes_processing>
    element, then the algorithm is applied to the union of all objects
    selected.
    The <objects_parameter>, <map_input> and <objectscond_extension> elements
    within the <objects_condition> element apply only to those objects selected
    in the condition. If an object is selected by more than one
    <objects_condition> element, it should receive parameter assignments and
    input mapping as if it were the first element so selected. These conditions
    include the identification and mapping of the variables for input::

    <!ELEMENT objects_condition (qticomment? ,
            (outcomes_metadata | and_objects | or_objects | not_objects)? ,
            objects_parameter* , map_input* , objectscond_extension?)>"""
    XMLNAME = 'objects_condition'
    XMLCONTENT = xml.ElementContent

    def __init__(self, parent):
        common.QTICommentContainer.__init__(self, parent)
        self.OutcomesOperator = None
        self.ObjectsParameter = []
        self.MapInput = []
        self.ObjectsCondExtension = None

    def get_children(self):
        for child in common.QTICommentContainer.get_children(self):
            yield child
        if self.OutcomesOperator:
            yield self.OutcomesOperator
        for child in itertools.chain(
                self.ObjectsParameter, self.MapInput):
            yield child
        if self.ObjectsCondExtension:
            yield self.ObjectsCondExtension


class ObjectsParameter(core.QTIElement):

    """This element contains the value of a particular parameter that is to be
    used by the corresponding scoring algorithm variable selection. Each
    parameter has a particular meaning to each scoring algorithm i.e. there is
    no established vocabulary for these parameters. These parameters are
    applied only to the objects selected by the enclosing <outcomes_condition>
    element.
    If multiple <outomes_condition> select the same object, the parameters are
    taken from the first <outcomes_condition> that selects the element::

    <!ELEMENT objects_parameter (#PCDATA)>
    <!ATTLIST objects_parameter  pname CDATA  #REQUIRED >"""
    XMLNAME = 'objects_parameter'
    XMLATTR_pname = 'pName'
    XMLCONTENT = xml.XMLMixedContent

    def __init__(self, parent):
        core.QTIElement.__init__(self, parent)
        self.pName = None


class MapInput(core.QTIElement):

    """This element is used to re-map the named input variable to another
    variable. The default variable names are derived from the type of scoring
    algorithm identified in <outcomes_processing> element. The target variable
    name must have been declared in the evaluation objects that undergoing
    aggregation and must be of the same type as the default variable. When a
    variable is remapped, all of its derived variables are to be remapped as
    well. The input mapping is applied only to the objects selected by the
    enclosing <outcomes_condition> element. If multiple <outcomes_condition>
    select the same object, the input mapping used is the one contained in the
    first <outcomes_condition> that selects the element::

    <!ELEMENT map_input (#PCDATA)>
    <!ATTLIST map_input  varname CDATA  'SCORE' >"""
    XMLNAME = 'map_input'
    XMLATTR_varname = 'varName'
    XMLCONTENT = xml.XMLMixedContent

    def __init__(self, parent):
        core.QTIElement.__init__(self, parent)
        self.varName = None


class ObjectsCondExtension(core.QTIElement):

    """This element contains the proprietary extensions that can be used to
    extend the functionally capabilities of the <outcomes_condition> element::

    <!ELEMENT objectscond_extension (#PCDATA)>"""
    XMLNAME = "objectscond_extension"
    XMLCONTENT = xml.XMLMixedContent


class OutcomesFeedbackTest(core.QTIElement):

    """The <outcomes_feedback> element contains the tests to be applied to
    determine if any and the type of feedback to be presented. This feedback
    could include information about passing the assessment etc::

    <!ELEMENT outcomes_feedback_test (test_variable , displayfeedback+)>
    <!ATTLIST outcomes_feedback_test  title CDATA  #IMPLIED >"""
    XMLNAME = 'outcomes_feedback_test'
    XMLATTR_title = 'title'
    XMLCONTENT = xml.ElementContent

    def __init__(self, parent):
        core.QTIElement.__init__(self, parent)
        self.title = None
        self.TestVariable = TestVariable(self)
        self.DisplayFeedback = []

    def get_children(self):
        yield self.TestVariable
        for child in self.DisplayFeedback:
            yield child


class OutcomesOperator(core.QTIElement):

    """Abstract class for::

    (outcomes_metadata | and_objects | or_objects | not_objects)"""
    pass


class OutcomesMetadata(OutcomesOperator):

    """This element defines the rule that is applied to the IMS QTI-specific
    and/or IMS Meta-data fields of the object to decide if the object scoring
    is to be aggregated. The content contains the value of the meta-data field
    that is being tested for within the rule. Data-type = String (1-64 chars)::

    <!ELEMENT outcomes_metadata (#PCDATA)>
    <!ATTLIST outcomes_metadata
            mdname CDATA  #REQUIRED
            mdoperator  (EQ | NEQ | LT | LTE | GT | GTE )  #REQUIRED >"""
    XMLNAME = 'outcomes_metadata'
    XMLATTR_mdname = 'mdName'
    XMLATTR_mdoperator = (
        'mdOperator',
        core.MDOperator.from_str_lower,
        core.MDOperator.to_str)
    XMLCONTENT = xml.ElementContent

    def __init__(self, parent):
        OutcomesOperator.__init__(self, parent)
        self.mdname = None
        self.mdoperator = None


class AndObjects(OutcomesOperator):

    """The <and_objects> element supports the construction of complex score
    condition rules to be built based upon the logical 'AND' operator. The
    object is selected for aggregation if all of the contained rules are
    'True'::

    <!ELEMENT and_objects (outcomes_metadata | and_objects | or_objects |
    not_objects)+>"""
    XMLNAME = 'and_objects'
    XMLCONTENT = xml.ElementContent

    def __init__(self, parent):
        OutcomesOperator.__init__(self, parent)
        self.OutcomesOperator = []

    def get_children(self):
        return iter(self.OutcomesOperator)


class OrObjects(OutcomesOperator):

    """The <or_objects> element supports the construction of complex score
    condition rules to be built based upon the logical 'OR' operator. The
    object is selected for aggregation if at least one of the contained rules
    is 'True'::

    <!ELEMENT or_objects (outcomes_metadata | and_objects | or_objects |
    not_objects)+>"""
    XMLNAME = 'or_objects'
    XMLCONTENT = xml.ElementContent

    def __init__(self, parent):
        OutcomesOperator.__init__(self, parent)
        self.OutcomesOperator = []

    def get_children(self):
        return iter(self.OutcomesOperator)


class NotObjects(OutcomesOperator):

    """The <not_objects> element supports the construction of complex rules to
    be built based upon the logical 'NOT' operator. The object is selected for
    aggregation if the contained rule(s) is 'False'::

    <!ELEMENT not_objects (outcomes_metadata | and_objects | or_objects |
    not_objects)>
    """
    XMLNAME = 'not_objects'
    XMLCONTENT = xml.ElementContent

    def __init__(self, parent):
        OutcomesOperator.__init__(self, parent)
        self.OutcomesOperator = None

    def get_children(self):
        if self.OutcomesOperator:
            yield self.OutcomesOperator


class TestVariable(core.QTIElement):

    """The <test_variable> element contains the conditions that are applied to a
    defined set of input scoring variables to determine if feedback is to be
    presented. Complex test structures can be constructed using the associated
    logic elements::

    <!ELEMENT test_variable (variable_test | and_test | or_test | not_test)>"""
    XMLNAME = 'test_variable'
    XMLCONTENT = xml.ElementContent

    def __init__(self, parent):
        core.QTIElement.__init__(self, parent)
        self.OutcomesTest = None

    def get_children(self):
        if self.OutcomesTest:
            yield self.OutcomesTest


class OutcomesTest(core.QTIElement):

    """Abstract class for::

    (variable_test | and_test | or_test | not_test)"""
    pass


class VariableTest(OutcomesTest):

    """The conditional test that is to be applied to the aggregated score
    variables. A wide range of separate and combinatorial tests can be
    applied::

    <!ELEMENT variable_test (#PCDATA)>
    <!ATTLIST variable_test
            varname CDATA  'SCORE'
            testoperator  (EQ | NEQ | LT | LTE | GT | GTE )  #REQUIRED >"""
    XMLNAME = 'variable_test'
    XMLATTR_varname = 'varName'
    XMLATTR_testoperator = (
        'testOperator',
        core.TestOperator.from_str_lower,
        core.TestOperator.to_str)
    XMLCONTENT = xml.XMLMixedContent

    def __init__(self, parent):
        OutcomesTest.__init__(self, parent)
        self.varName = None
        self.testOperator = None


class AndTest(OutcomesTest):

    """The <and_test> element is used to define variable test conditions all of
    which must be defined as 'True' to result in the output variables being
    assigned the corresponding values::

    <!ELEMENT and_test (variable_test | and_test | or_test | not_test)+>"""
    XMLNAME = 'and_test'
    XMLCONTENT = xml.ElementContent

    def __init__(self, parent):
        OutcomesTest.__init__(self, parent)
        self.OutcomesTest = []


class OrTest(OutcomesTest):

    """The <or_test> element is used to define variable test conditions at least
    one of which must be defined as 'True' to result in the output variables
    being assigned the corresponding values::

    <!ELEMENT or_test (variable_test | and_test | or_test | not_test)+>"""
    XMLNAME = 'or_test'
    XMLCONTENT = xml.ElementContent

    def __init__(self, parent):
        OutcomesTest.__init__(self, parent)
        self.OutcomesTest = []


class NotTest(OutcomesTest):

    """The <not_test> element is used to define variable test conditions that
    result in the condition being defined as 'False' to result in the output
    variables being assigned the corresponding values::

    <!ELEMENT not_test (variable_test | and_test | or_test | not_test)>"""
    XMLNAME = 'not_test'
    XMLCONTENT = xml.ElementContent

    def __init__(self, parent):
        OutcomesTest.__init__(self, parent)
        self.OutcomesTest = None



#! /usr/bin/env python
"""This module implements the QTI 1.2.1 specification defined by IMS GLC"""

import warnings

from .qtiv1.assessment import *     # noqa
from .qtiv1.common import *         # noqa
from .qtiv1.core import *           # noqa
from .qtiv1.item import *           # noqa
from .qtiv1.objectbank import *     # noqa
from .qtiv1.outcomes import *       # noqa
from .qtiv1.sao import *            # noqa
from .qtiv1.section import *        # noqa
from .qtiv1.xml import *            # noqa


warnings.warn(
    "pyslet.imsqtiv1p2p1 is deprecated, use pyslet.qtiv1 package instead",
    DeprecationWarning, stacklevel=3)



#! /usr/bin/env python
"""OData core elements"""

import warnings

from .. import rfc4287 as atom
from ..http import grammar
from ..http import params
from ..pep8 import MigratedClass, old_method
from ..xml import namespace as xmlns

from . import csdl as edm
from . import edmx
from . import core


# Legacy name for compatibilty
InvalidMetadataDocument = edm.InvalidMetadataDocument


class FeedCustomisationMixin(MigratedClass):

    """Utility class used to add common feed customisation attributes"""

    AtomPaths = {
        'SyndicationAuthorName': [
            (atom.ATOM_NAMESPACE, "author"),
            (atom.ATOM_NAMESPACE, "name")],
        'SyndicationAuthorEmail': [
            (atom.ATOM_NAMESPACE, "author"),
            (atom.ATOM_NAMESPACE, "email")],
        'SyndicationAuthorUri': [
            (atom.ATOM_NAMESPACE, "author"),
            (atom.ATOM_NAMESPACE, "uri")],
        'SyndicationPublished': [(atom.ATOM_NAMESPACE, "published")],
        'SyndicationRights': [(atom.ATOM_NAMESPACE, "rights")],
        'SyndicationSummary': [(atom.ATOM_NAMESPACE, "summary")],
        'SyndicationTitle': [(atom.ATOM_NAMESPACE, "title")],
        'SyndicationUpdated': [(atom.ATOM_NAMESPACE, "updated")],
        'SyndicationContributorName': [
            (atom.ATOM_NAMESPACE, "contributor"),
            (atom.ATOM_NAMESPACE, "name")],
        'SyndicationContributorEmail': [
            (atom.ATOM_NAMESPACE, "contributor"),
            (atom.ATOM_NAMESPACE, "email")],
        'SyndicationContributorUri': [(atom.ATOM_NAMESPACE, "source")]
    }

    @old_method('GetTargetPath')
    def get_target_path(self):
        """Returns the target path for an element

        The result is a list of qualified element names, that is, tuples
        of (namespace,name). The last name may start with '@' indicating
        an attribute rather than an element.

        Feed customisations are declared using the FC_TargetPath
        attribute.  Returns None if there is no target path declared."""
        try:
            path = self.get_attribute(core.FC_TargetPath)
            if path in self.AtomPaths:
                return self.AtomPaths[path]
            else:
                ns = self.get_attribute(core.FC_NsUri)
                return [(ns, x) for x in path.split('/')]
        except KeyError:
            return None

    @old_method('KeepInContent')
    def keep_in_content(self):
        """Returns true if a property value should be kept in the content

        This is indicated with the FC_KeepInContent attribute.  If the
        attribute is missing then False is returned, so properties with
        custom paths default to being omitted from the properties
        list."""
        try:
            return self.get_attribute(core.FC_KeepInContent) == "true"
        except KeyError:
            return False

    @old_method('GetFCNsPrefix')
    def get_fc_ns_prefix(self):
        """Returns the custom namespace mapping to use.

        The value is read from the FC_NsPrefix attribute.  The result is
        a tuple of: (prefix, namespace uri).

        If no mapping is specified then (None,None) is returned."""
        try:
            prefix = self.get_attribute(core.FC_NsPrefix)
            ns = self.get_attribute(core.FC_NsUri)
            return prefix, ns
        except KeyError:
            return None, None


class EntityType(edm.EntityType, FeedCustomisationMixin):

    """Supports feed customisation behaviour of EntityTypes"""

    @old_method('GetSourcePath')
    def get_source_path(self):
        """Returns the source path

        This result is read from the FC_SourcePath attribute.  It is a
        *list* of property names that represents a path into the entity
        or None if there is no source path set."""
        try:
            return self.get_attribute(core.FC_SourcePath).split('/')
        except KeyError:
            return None

    @old_method('HasStream')
    def has_stream(self):
        """Returns true if this is a media link resource.

        Read from the HasStream attribute.  The default is False."""
        try:
            return self.get_attribute(core.HAS_STREAM) == "true"
        except KeyError:
            return False


class Property(edm.Property, FeedCustomisationMixin):

    """Supports feed customisation behaviour of Properties"""

    @old_method('GetMimeType')
    def get_mime_type(self):
        """Returns the media type of a property

        The result is read from the MimeType attribute.  It is a
        :py:class:`~pyslet.http.params.MediaType` instance or None if
        the attribute is not defined."""
        try:
            return params.MediaType.from_str(
                self.get_attribute(core.MIME_TYPE))
        except KeyError:
            return None

    def __call__(self, literal=None):
        """Overridden to add mime type handling

        Property elements are callable in the core EDM, returning an
        :py:class:`~pyslet.odata2.core.EDMValue` object instantiated
        from the declaration.  This implementation adds to the base
        behaviour by reading the optional mime type attribute and adding
        it to the value if applicable."""
        value = super(Property, self).__call__(literal)
        value.mtype = self.get_mime_type()
        return value


class EntityContainer(edm.EntityContainer):

    """Supports OData's concept of the default container."""

    @old_method('IsDefaultEntityContainer')
    def is_default_entity_container(self):
        """Returns True if this is the default entity container

        The value is read from the IsDefaultEntityContainer attribute.
        The default is False."""
        try:
            return self.get_attribute(
                core.IS_DEFAULT_ENTITY_CONTAINER) == "true"
        except KeyError:
            return False

    def content_changed(self):
        super(EntityContainer, self).content_changed()
        if self.is_default_entity_container():
            ds = self.find_parent(DataServices)
            if ds is not None:
                ds.defaultContainer = self


class EntitySet(edm.EntitySet):

    def set_location(self):
        """Overridden to add support for the default entity container

        By default, the path to an EntitySet includes the name of the
        container it belongs to, e.g., MyDatabase.MyTable.  This
        implementation checks to see if we in the default container and,
        if so, omits the container name prefix before setting the
        location URI."""
        container = self.find_parent(EntityContainer)
        if container and not container.is_default_entity_container():
            path = container.name + '.' + self.name
        else:
            path = self.name
        self.location = self.resolve_uri(path)


class DataServices(edmx.DataServices):

    """Adds OData specific behaviour"""

    def __init__(self, parent):
        super(DataServices, self).__init__(parent)
        #: the default entity container
        self.defaultContainer = None

    @old_method('DataServicesVersion')
    def data_services_version(self):
        """Returns the data service version

        Read from the DataServiceVersion attribute.  Defaults to None."""
        try:
            return self.get_attribute(core.DATA_SERVICE_VERSION)
        except KeyError:
            return None

    @old_method('SearchContainers')
    def search_containers(self, name):
        """Returns an entity set or service operation with *name*

        *name* must be of the form::

            [<entity container>.]<entity set, function or operation name>

        The entity container must be present unless the target is in the
        default container in which case it *must not* be present.

        If *name* can't be found KeyError is raised."""
        resource = None
        if name in self.defaultContainer:
            return self.defaultContainer[name]
        else:
            for s in self.Schema:
                if name in s:
                    resource = s[name]
                    container = resource.find_parent(edm.EntityContainer)
                    if container is self.defaultContainer:
                        continue
                    return resource
        raise KeyError(
            "No entity set or service operation with name %s" % name)


class Edmx(edmx.Edmx):

    """The root element of OData-specific metadata documents"""

    DataServicesClass = DataServices


def ValidateMetadataDocument(doc):      # noqa
    warnings.warn(
        "ValidateMetadataDocument is deprecated, use validate method instead",
        DeprecationWarning,
        stacklevel=3)
    return doc.validate()


class Document(edmx.Document):

    """Class for working with OData-specific metadata documents.

    Adds namespace prefix declarations for the OData metadata and OData
    dataservices namespaces."""

    classMap = {}

    def __init__(self, **args):
        edmx.Document.__init__(self, **args)
        self.make_prefix(core.ODATA_METADATA_NAMESPACE, 'm')
        self.make_prefix(core.ODATA_DATASERVICES_NAMESPACE, 'd')

    @classmethod
    def get_element_class(cls, name):
        """Returns the class used to represent an element.

        Overrides
        :py:meth:`~pyslet.odata2.edmx.Document.get_element_class` to use
        the OData-specific implementations of the edmx/csdl classes
        defined in this module."""
        result = Document.classMap.get(name, None)
        if result is None:
            result = edmx.Document.get_element_class(name)
        return result

    @old_method('Validate')
    def validate(self):
        """Validates any declared OData extensions

        Checks many of the requirements given in the specification and
        raises :py:class:`~pyslet.odata2.csdl.InvalidMetadataDocument`
        if the tests fail.

        Returns the data service version required to process the service
        or None if no data service version is specified."""
        super(Document, self).validate()
        # IsDefaultEntityContainer: This attribute MUST be used on an
        # EntityContainer element to indicate which EntityContainer is the
        # default container for the data service. Each conceptual schema
        # definition language (CSDL) document used to describe a data
        # service MUST mark exactly one EntityContainer with this attribute
        # to denote it is the default.
        ndefaults = 0
        for container in self.root.find_children_depth_first(
                edm.EntityContainer):
            try:
                flag = container.get_attribute(
                    core.IS_DEFAULT_ENTITY_CONTAINER)
                if flag == "true":
                    ndefaults += 1
                elif flag != "false":
                    raise edm.InvalidMetadataDocument(
                        "IsDefaultEntityContainer: %s" % flag)
            except KeyError:
                pass
        if ndefaults != 1:
            raise edm.InvalidMetadataDocument(
                "IsDefaultEntityContainer required on "
                "one and only one EntityContainer")
        for p in self.root.find_children_depth_first(edm.Property):
            try:
                params.MediaType.from_str(p.get_attribute(core.MIME_TYPE))
            except grammar.BadSyntax as e:
                raise edm.InvalidMetadataDocument(
                    "MimeType format error in property %s: %s" %
                    (p.name, str(e)))
            except KeyError:
                pass
        # HttpMethod: This attribute MUST be used on a <FunctionImport>
        # element to indicate the HTTP method which is to be used to invoke
        # the ServiceOperation exposing the FunctionImport
        for f in self.root.find_children_depth_first(edm.FunctionImport):
            try:
                if f.get_attribute(core.HttpMethod) in (
                        "POST", "PUT", "GET", "MERGE", "DELETE"):
                    continue
                raise edm.InvalidMetadataDocument(
                    "Bad HttpMethod: %s" % f.get_attribute(core.HttpMethod))
            except KeyError:
                raise edm.InvalidMetadataDocument(
                    "FunctionImport must have HttpMethod defined: %s" % f.name)
        # HasStream: This attribute MUST only be used on an <EntityType>
        # element
        for e in self.root.find_children_depth_first(edm.CSDLElement):
            try:
                hs = e.get_attribute(core.HAS_STREAM)
                if not isinstance(e, edm.EntityType):
                    raise edm.InvalidMetadataDocument(
                        "HasStream must only be used on EntityType")
                elif hs not in ("true", "false"):
                    raise edm.InvalidMetadataDocument(
                        "Bad value for HasStream: %s" % hs)
            except KeyError:
                pass
        # DataServiceVersion: This attribute MUST be in the data service
        # metadata namespace and SHOULD be present on a <edmx:DataServices>
        # element to indicate the version of the data service CSDL
        # annotations (attributes in the data service metadata namespace)
        # used by the document.
        #
        # The value of this attribute MUST be 1.0 unless a
        # "FC_KeepInContent" Customizable Feed annotation with a value equal
        # to false is present... In this case, the attribute value MUST be
        # 2.0.
        try:
            version = self.root.DataServices.get_attribute(
                core.DATA_SERVICE_VERSION)
            match = "1.0"
            for e in self.root.find_children_depth_first(edm.CSDLElement):
                try:
                    if e.get_attribute(core.FC_KeepInContent) == "false":
                        match = "2.0"
                        break
                except KeyError:
                    pass
            if version != match:
                raise edm.InvalidMetadataDocument(
                    "Expected version %s, found %s" % (match, version))
            return version
        except KeyError:
            return None

xmlns.map_class_elements(Document.classMap, globals(), edm.NAMESPACE_ALIASES)



#! /usr/bin/env python
"""Implements the Atom Publishing Protocol defined in RFC 5023

References:

IRIs [RFC3987]; cf URI [RFC3986]
Before an IRI in a document is used by HTTP, the IRI is first converted to a
URI according to the procedure defined in Section 3.1 of [RFC3987]

xml:base attribute [W3C.REC-xmlbase-20010627]
xml:lang attribute [W3C.REC-xml-20040204], Section 2.12
"""

import sys
import io
import traceback

from . import pep8
from . import rfc2396 as uri
from . import rfc4287 as atom
from .http import client as http
from .py2 import (
    byte,
    byte_value,
    character,
    is_text,
    SortableMixin,
    to_text,
    uempty,
    UnicodeMixin)
from .xml import structures as xml
from .xml import namespace as xmlns


#: The namespace to use for Atom Publishing Protocol elements
APP_NAMESPACE = "http://www.w3.org/2007/app"
#: The mime type for service documents
ATOMSVC_MIMETYPE = "application/atomsvc+xml"
#: The mime type for category documents
ATOMCAT_MIMETYPE = "application/atomcat+xml"

APP_MIMETYPES = {
    ATOMSVC_MIMETYPE: True,
    ATOMCAT_MIMETYPE: True,
    atom.ATOM_MIMETYPE: True
}


def parse_yes_no(src):
    return src.strip().lower() == 'yes'


def format_yes_no(value):
    if value:
        return 'yes'
    else:
        return 'no'


class APPElement(xmlns.XMLNSElement):

    """Base class for all APP elements.

    All APP elements can have xml:base, xml:lang and/or xml:space attributes.
    These are handled by the base
    :py:class:`~pyslet.xml.structures.Element` base class."""
    pass


class Accept(APPElement):

    """Represents the accept element."""
    XMLNAME = (APP_NAMESPACE, 'accept')


class Categories(APPElement):

    """The root of a Category Document.

    A category document is a document that describes the categories
    allowed in a collection."""
    XMLNAME = (APP_NAMESPACE, 'categories')

    XMLATTR_href = ('href', uri.URI.from_octets, str)
    XMLATTR_fixed = ('fixed', parse_yes_no, format_yes_no)
    XMLATTR_scheme = 'scheme'
    XMLCONTENT = xml.ElementContent

    def __init__(self, parent):
        APPElement.__init__(self, parent)
        self.href = None
        #: an optional :py:class:`~pyslet.rfc2396.URI` to the category
        self.fixed = None
        #: indicates whether the list of categories is a fixed set. By
        #: default they're open.
        self.scheme = None
        #: identifies the default scheme for categories defined by this
        #: element
        self.Category = []
        #: the list of categories, instances of
        #: :py:class:~pyslet.rfc4287.Category

    def get_children(self):  # noqa
        for child in self.Category:
            yield child
        for child in APPElement.get_children(self):
            yield child


class Service(APPElement):

    """The container for service information

    Associated with one or more Workspaces."""
    XMLNAME = (APP_NAMESPACE, 'service')
    XMLCONTENT = xml.ElementContent

    def __init__(self, parent):
        APPElement.__init__(self, parent)
        self.Workspace = []		#: a list of :py:class:`Workspace` instances

    def get_children(self):  # noqa
        for child in APPElement.get_children(self):
            yield child
        for child in self.Workspace:
            yield child


class Workspace(APPElement):

    """Workspaces are server-defined groups of Collections."""
    XMLNAME = (APP_NAMESPACE, 'workspace')
    XMLCONTENT = xml.ElementContent

    def __init__(self, parent):
        APPElement.__init__(self, parent)
        self.Title = None			#: the title of this workspace
        self.Collection = []		#: a list of :py:class:`Collection`

    def get_children(self):  # noqa
        for child in APPElement.get_children(self):
            yield child
        if self.Title:
            yield self.Title
        for child in self.Collection:
            yield child


class Collection(APPElement):

    """Describes a collection (feed)."""
    XMLNAME = (APP_NAMESPACE, 'collection')
    XMLCONTENT = xml.ElementContent

    XMLATTR_href = ('href', uri.URI.from_octets, str)

    def __init__(self, parent):
        APPElement.__init__(self, parent)
        self.href = None
        #: the URI of the collection (feed)
        self.Title = None
        #: the human readable title of the collection
        self.Accept = []
        #: list of :py:class:`Accept` media ranges that can be posted to
        #: the collection
        self.Categories = []
        # : list of :py:class:`Categories` that can be applied to
        # : members of the collection

    def get_children(self):  # noqa
        for child in APPElement.get_children(self):
            yield child
        if self.Title:
            yield self.Title
        for child in self.Accept:
            yield child
        for child in self.Categories:
            yield child

    @pep8.old_method('GetFeedURL')
    def get_feed_url(self):
        """Returns a fully resolved URL for the collection (feed)."""
        return self.resolve_uri(self.href)


class Document(atom.AtomDocument):

    """Class for working with APP documents.

    This call can represent both APP and Atom documents."""
    classMap = {}

    def __init__(self, **args):
        atom.AtomDocument.__init__(self, **args)
        self.defaultNS = APP_NAMESPACE

    def ValidateMimeType(self, mimetype):   # noqa
        """Checks *mimetype* against the APP or Atom specifications."""
        return (mimetype in APP_MIMETYPES or
                atom.AtomDocument.ValidateMimeType(self, mimetype))

    @classmethod
    def get_element_class(cls, name):
        """Returns the APP or Atom class used to represent name.

        Overrides
        :py:meth:`~pyslet.rfc4287.AtomDocument.get_element_class` when the
        namespace is :py:data:`APP_NAMESPACE`."""
        if name[0] == APP_NAMESPACE:
            return cls.classMap.get(
                name, atom.AtomDocument.classMap.get((name[0], None),
                                                     APPElement))
        else:
            return atom.AtomDocument.get_element_class(name)

xmlns.map_class_elements(Document.classMap, globals())


class Slug(UnicodeMixin, SortableMixin):

    """Represents an HTTP slug header value.

    slug
            The opaque slug as a unicode string

    The built-in str function can be used to format instances according
    to the grammar defined in the specification.

    Instances must be treated as immutable, they define comparison
    methods and a hash implementation to allow them to be used as keys
    in dictionaries."""

    def __init__(self, slug):
        self.slug = slug		#: the slug value

    @classmethod
    def from_str(cls, source):
        """Creates a slug from a *source* string."""
        return cls(uri.unescape_data(source).decode('utf-8'))

    escape_byte = byte('%')

    def __unicode__(self):
        result = []
        for c in self.slug.encode('utf-8'):
            cv = byte_value(c)
            if c == self.escape_byte or cv < 0x20 or cv > 0x7E:
                result.append("%%%02X" % cv)
            else:
                result.append(character(c))
        return uempty.join(result)

    def __repr__(self):
        return "Slug(%s)" % repr(self.slug)

    def sortkey(self):
        return self.slug

    def otherkey(self, other):
        if is_text(other):
            other = self.from_str(other)
        if isinstance(other, self.__class__):
            return other.sortkey()
        else:
            return NotImplemented

    def __hash__(self):
        return hash(self.slug)


class Client(http.Client):

    def __init__(self, **kwargs):
        http.Client.__init__(self, **kwargs)

    def queue_request(self, request, timeout=60):
        # if there is no Accept header, add one
        if not request.has_header('Accept'):
            request.set_header('Accept', ','.join(
                (atom.ATOM_MIMETYPE, ATOMSVC_MIMETYPE, ATOMCAT_MIMETYPE)),
                True)
        super(Client, self).queue_request(request, timeout)


class Server(pep8.PEP8Compatibility):

    def __init__(self, service_root='http://localhost/', **kwargs):
        pep8.PEP8Compatibility.__init__(self)
        service_root = kwargs.get('serviceRoot', service_root)
        if not isinstance(service_root, uri.URI):
            #: the canonical URL of the service root
            self.service_root = \
                uri.URI.from_octets(service_root).canonicalize()
        else:
            self.service_root = service_root.canonicalize()
        #: the :py:class:`Service` instance that describes this service.
        self.serviceDoc = Document(root=Service, base_uri=self.service_root)
        self.service = self.serviceDoc.root
        # make the base explicit in the document
        self.service.set_base(str(self.service_root))
        #: set this to True to expose python tracebacks in 500
        #: responses, defaults to False
        self.debugMode = False

    def __call__(self, environ, start_response):
        """wsgi interface for calling instances of this Atom server object.

        We add an additional optional parameter *response_headers*"""
        response_headers = []
        if environ['SCRIPT_NAME'] + environ['PATH_INFO'] == \
                self.service_root.abs_path:
            data = to_text(self.serviceDoc).encode('utf-8')
            response_headers.append(("Content-Type", ATOMSVC_MIMETYPE))
            response_headers.append(("Content-Length", str(len(data))))
            start_response("200 Ok", response_headers)
            return [data]
        else:
            return self.HandleMissing(environ, start_response)

    def handle_missing(self, environ, start_response):
        response_headers = []
        data = "This server supports the Atom Publishing Protocol\r\n"\
            "For service information see: %s" % str(self.service_root)
        response_headers.append(("Content-Length", str(len(data))))
        response_headers.append(("Content-Type", 'text/plain'))
        start_response("404 Not found", response_headers)
        return [data]

    def handle_error(self, environ, start_response, code=500):
        """Designed to be called by an otherwise uncaught exception.

        Generates a 500 response by default."""
        response_headers = []
        cdata = io.BytesIO()
        if self.debugMode:
            traceback.print_exception(*sys.exc_info(), file=cdata)
        else:
            cdata.write("Sorry, there was an internal error "
                        "while processing this request")
        response_headers.append(("Content-Type", 'text/plain'))
        response_headers.append(("Content-Length", str(len(cdata.getvalue()))))
        start_response("%i Unexpected error" % code, response_headers)
        return [cdata.getvalue()]
